# from genet.utils import *
import os, sys, regex
import pandas as pd
from Bio.Seq import reverse_complement, translate
from Bio.SeqUtils import gc_fraction
from genet.design.DesignUtils import dict_pam_disrup_rank, test_score_data

'''
TODO
1. flash를 python code로 구현한 것이 없으므로, 여기서 input은 .fastq 파일만 가능
2. 나중에 flashpy 개발이 좀 더 진행되면 도입하는 것을 생각해보자
3. python 기본적으로 python 3.6~3.10까지 호환되는 것을 목표로 하고, 3.11도 테스트하기

'''


class MakeStop:
    '''
    특정 gene symbol을 넣으면,
    그 gene에서 SNV (1bp substitution)으로 만들 수 있는 모든 stop codon의 경우의 수를 return 한다.

    input  = gene symbol (ensembl 기준)
    output = DataFrame (gene | pos | WT (121bp) | ED (121bp))

    example:
    >>> from genet.design import MakeStop
    >>> df_out = MakeStop('ACE2')

    '''
    def __init__(self, gene_name:str):
        print('Start MakeStop')

# class END: MakeStop



class SNVGenerator:

    def __init__(self, seq:str, pos:int) -> list:
        """sequence에서 지정된 위치에 A/T/G/C로 만들어진 SNV를 만들어서 list로 return하는 함수

        Args:
            * seq (str): SNV를 만들 서열 정보
            * pos (int): seq에서 SNV를 만들 위치 정보

        Returns:
            list: 특정 위치에 가능한 SNV들이 모인 list

        Eaxmple
            >>> from genet.design import MakeSNVs
            >>> seq = 'ATGGCTGACTGC'
            >>> list_snvs = MakeSNVs(seq, 4)

            list_snvs = ['ATGGATGACTGC', 'ATGGGTGACTGC', 'ATGGTTGACTGC']
        """        

        list_base = ['A', 'G', 'C', 'T']
        self.list_sSNV = []

        for base in list_base:
            list_sSeq = list(seq)
            list_sSeq[pos] = base
            snv_temp = ''.join(list_sSeq)
            if seq!=snv_temp: self.list_sSNV.append(snv_temp)

# class END: MakeSNVs


class SynonymousPE:

    def __init__(self,
                 dp_record:pd.Series,
                 ref_seq:str,
                 frame:int,
                 cds_start:int=0,
                 cds_end:int=121,
                 adj_rha:bool=True,
                 mut_target:str=None,
                 mut_in_cds_filter:bool = False
                 ):
        """DeepPrime output으로 만들어진 파일에서 pegRNA들에 silent mutation을 함께 유발하는 것 중 최적의 pegRNA를 선택해주는 것
        모든 기능은 prime editing으로 1bp substitution을 했을 때를 가정하여 만들어졌다.
        우선, intended edit 기준으로 best pegRNA들을 선정한다.
        그 후, 아래와 같이 silent PAM co-editing을 추가한다.

        Case 1: Silent PAM co-editing이 LHA 부분에 발생하는 경우,
            그대로 silent PAM co-editing을 넣고 완성

        Case 2: Silent PAM co-editing이 RHA 부분에 발생하는 경우,
            원래 pegRNA의 RHA 길이를 기준으로 맞춰주기 위해 RTT 길이를 늘려주기.
        
        Case 3: Intended edit이 이미 PAM 위치 (+5 or +6)인 경우,
        1) 만약 남은 PAM 위치 (G)에 silent mutation이 가능하면 그 것으로 결정
        2) 남은 PAM 위치의 SNV 중 silent mutation이 없다면, LHA에서 가능한 silent mutation을 넣어준다.
        3) 만약 선택되는 mutation position이 splicing adaptor 위치라면, (cds_start / end 앞뒤로 5nt) 차순위로 뽑기
        4) 차순위로 불가능한 것들이 있다면, RHA에서 silent mutation을 뽑는다.         
    
        Args:
            * dp_record (pd.Series): DeepPrime을 돌리고 output으로 나오는 것의 한 index의 정보를 가져온 것.
            * ref_seq (str): 해당 mutation을 만들었을 떄, 사용한 reference sequence (121nt)
            * frame (int): Ref_seq의 frame을 표시 (0, 1, 2)
            * cds_start (int, optional): CDS가 시작되는 위치, 그 이전 위치에서부터는 silent mutation을 만들 수 없으므로 고르는 위치에서 제외. Defaults to 0.
            * cds_end (int, optional): CDS가 종료되는 위치, 그 이후 위치에서부터는 silent mutation을 만들 수 없으므로 고르는 위치에서 제외. Defaults to 121.
            * adj_rha (bool, optional): silent mutation이 RHA에 위치하는 경우, 기존의 pegRNA에서의 RHA 길이만큼을 유지하기 위해 RTT를 늘려주는 기능. Defaults to True.
            * mut_target (str, optional): Synonymous mutation을 특정 패턴의 sequence에서 찾아서 하고 싶을 경우. 해당 sequence가 들어있는 codon에 만든다.
            * mut_in_cds_filter (bool): synonymous mutation을 coding sequence (CDS)에만 넣을지를 의미함. (True/False)

        Raises:
            ValueError: frame이 0, 1, 2 중에 하나로 입력되지 않은 경우
            ValueError: Reference sequence가 pegRNA와 matching이 안 될 경우
        """        
    
        # input error check
        if type(dp_record) != type(pd.Series(dtype = 'object')): raise TypeError("The type of 'dp_record' should be pd.Series.")
        if frame not in [0, 1, 2]              : raise ValueError('Frame should be 0, 1 or 2')

        # step 1: Get required features
        '''
        cds_start / end / frame 등을 variants마다 각각 따로 넣어줘야 하는 것이 복잡한 것 같음
        Exon의 길이 (fixed), exon의 frame, SNV의 position만 넣으면, 
        각 SNV의 frame, CDS start, CDS end를 내부적으로 만들어서 계산하는 것으로 수정하기

        예를 들어, cds_end는 다음과 같이 구할 수 있음
        cds_end = 66 + exon_len - pos 
        if cds_end > 121: cds_end=121
        '''

        self.rec       = dp_record
        self.sID       = dp_record.ID
        self.rtpbs_dna = reverse_complement(dp_record['RT-PBS'])
        self.pbs_dna   = self.rtpbs_dna[:dp_record.PBS_len]
        self.rtt_dna   = self.rtpbs_dna[-dp_record.RTT_len:]
        self.rtt_len   = dp_record.RTT_len

        self.wt_seq    = dp_record.Target # WT74 seq
        self.ed_seq    = self.wt_seq[:21] + self.rtt_dna + self.wt_seq[21+self.rtt_len:] # ED74 seq

        self.edit_pos  = dp_record.Edit_pos
        self.ref_seq   = ref_seq.upper()
        self.frame     = frame
        self.cds_start = cds_start
        self.cds_end   = cds_end
        self.spacer    = dp_record.Spacer
        
        self.splicing_adaptor = [i for i in range(cds_start-5, cds_start)] + [i for i in range(cds_end, cds_end+5)]

        # step 2: pegRNA의 strand 방향에 따라 synonymous Mut 생성 함수 결정
        if self.wt_seq in self.ref_seq:
            self.strand = '+'
            #self.rtt_frame = (frame - self.edit_pos + 1) % 3 # rtt 시작점의 frame, LHA 길이를 이용한 계산
            self.rtt_frame = (frame - (self.edit_pos - 1) % 3) % 3 #frame - (self.edit_pos - 1)가 음수가 되었을 때 결과 값이 다름.

        elif reverse_complement(self.wt_seq) in self.ref_seq:
            self.strand = '-'
            #self.rtt_frame = (self.edit_pos + frame) % 3  # revcom_rtt_dna의 3' end가 위치하는 지점의 frame. 시작점이 거기이기 때문.
            self.rtt_frame = (self.edit_pos + frame - 1) % 3
            
        else: raise ValueError('Reference sequence is not matched with pegRNA information!\nPlease chech your ref_seq')
    
        self.dict_mut = {
            'Codon_WT'      : [],
            'Codon_Mut'     : [],
            'Codon_MutPos'  : [],
            'Codon_RefStart': [],
            'Codon_RefEnd'  : [],
            'RTT_DNA_frame' : [],
            'RTT_DNA_Strand': [],
            'AminoAcid_WT'  : [],
            'AminoAcid_Mut' : [],
            'Silent_check'  : [],
            'Mut_pos'       : [],
            'Mut_refpos'    : [],
            'PAM_Mut'       : [],
            'Priority'      : [],
            'Edit_class'    : [],
            'RTT_DNA_Mut'   : [],
            'Syn_in_CDS'    : [],
        }
        
        self.output = self.generate(self.rtt_frame, self.strand, mut_in_cds_filter)
        
        if type(self.output) == pd.core.series.Series:
            # SynonyPE가 만들어진 경우

            # step 3: 만약 RHA 길이 조정 옵션이 True로 되어있으면, 조정해주기. (defualt)
            if adj_rha == True:
                adj_len = self.output['Mut_pos'] - self.edit_pos
                
                if adj_len > 0: 
                    rtt_end = 21 + self.rtt_len
                    self.output['RTT_DNA_Mut'] = self.output['RTT_DNA_Mut'] + self.wt_seq[rtt_end:rtt_end+adj_len]

            self.extension = self.pbs_dna + self.output['RTT_DNA_Mut']
            
        elif type(self.output) == pd.core.frame.DataFrame:
            # SynonyPE가 안 만들어져서
            # output이 self.mutations로 나온 경우
            
            self.extension = 'Not available SynonyPE'
        
    # End def __init__:

        
    def _make_snv(self, seq:str, pos:int) -> list:
        """sequence에서 지정된 위치에 A/T/G/C로 만들어진 SNV를 만들어서 list로 return하는 함수

        Args:
            seq (str): SNV를 만들 서열 정보
            pos (int): seq에서 SNV를 만들 위치 정보

        Returns:
            list: 특정 위치에 가능한 SNV들이 모인 list
        """        

        list_base = ['A', 'G', 'C', 'T']
        list_sSNV = []

        for base in list_base:
            list_sSeq = list(seq)
            list_sSeq[pos] = base
            snv_temp = ''.join(list_sSeq)
            if seq!=snv_temp: list_sSNV.append(snv_temp)

        return list_sSNV
    # def END: _make_snv

    def generate(self, rtt_frame:int, strand:str, mut_in_cds_filter:bool) -> pd.DataFrame:

        """우선 만들 수 있는 모든 종류의 mutation table을 만든 다음, 각 mutation 마다 synonymous 여부, mut_type 등을 분류해준다.
        그리고 그 분류 기준을 우선순위에 따라 우선도 (priority)를 계산해준 다음, 정해진 기준에 따라 최적의 mutation을 선정해준다.
        
        Args:
            rtt_frame (int): CDS에서 RTT의 frame을 의미함 (0, 1, 2).
            strand (str): Reference sequence 기준으로 pegRNA의 방향 (+ / -)
            mut_in_cds_filter (bool): synonymous mutation을 coding sequence (CDS)에만 넣을지를 의미함 (True/False)

        Returns:
            pd.DataFrame: Mutation 정보들이 담긴 DataFrame.
        """

        ep = self.edit_pos

        ### 이 부분이 가장 중요한 부분 ########################################################
        ### 나중에 le / re 라는 표현보다 nick 방향에 따른 표현으로 바꾸기... 헷갈리는 표현이라서..

        if strand == '+':
            codon_le  = rtt_frame
            #codon_re  = (3 - (rtt_frame + self.rtt_len)) % 3
            codon_re  = -(rtt_frame + self.rtt_len) % 3

            codon_RTT = self.ed_seq[21-codon_le:21+self.rtt_len+codon_re]

            dict_codon_Pos = {codon_RTT: [i for i in range(codon_le, len(codon_RTT)-codon_re)]}

        else:
            #codon_re  = (self.frame - ep) % 3
            codon_re  = 2 - rtt_frame
            codon_le  = -(codon_re + self.rtt_len) % 3

            codon_RTT = self.ed_seq[21-codon_re:21+self.rtt_len+codon_le]
            codon_RTT = reverse_complement(codon_RTT)
            
            dict_codon_Pos = {codon_RTT: [i for i in range(codon_le, len(codon_RTT) - codon_re)]} #due to revcom sequence

        ### 결과가 안 맞으면, 여기까지 (codon_RTT)를 다시 한번 살펴보기, 특히 (-) strand! #######
        
        self.codon_le = codon_le
        self.codon_re = codon_re
        self.codon_RTT = codon_RTT

        # for loop: LHA 위치가 걸쳐있는 codon들을 가져온다.
        for codon in dict_codon_Pos:
            # for loop: 각 codon들에서 LHA에 해당하는 위치들을 가져온다.
            for snv_pos in dict_codon_Pos[codon]:

                if strand == '+': 
                    mut_pos = snv_pos + 1 - codon_le
                else: 
                    mut_pos = len(codon_RTT) - codon_re - snv_pos

                if mut_pos == ep: continue
                if mut_pos <  1 : continue

                list_mut_codon = self._make_snv(codon, snv_pos)

                
                for mut_codon in list_mut_codon:

                    if strand == '+':
                        rtt_dna_mut = mut_codon[codon_le:len(mut_codon)-codon_re]
                        mut_refpos  = 60 + (mut_pos - ep)
                        codon_start = 60 - (codon_le + ep - 1)
                        codon_end   = codon_start + len(codon_RTT)
                        
                    else:
                        rtt_dna_mut = reverse_complement(mut_codon[codon_le:len(mut_codon)-codon_re])
                        mut_refpos  = 60 - (mut_pos - ep)
                        codon_end   = 60 + (codon_re + ep - 1)
                        codon_start = codon_end - len(codon_RTT) + 1

                    # priority 결정하는 부분 ##########################################
                    # 1/ Edit class에 따라서 분류하고, 각 class에 따라 값을 할당
                    if (mut_pos in [5, 6]) and (rtt_dna_mut[4:6] not in ['GG', 'GA', 'AG']): 
                        if mut_pos < ep : edit_class = 'PAM_edit_LHA'; priority = 1
                        else            : edit_class = 'PAM_edit_RHA'; priority = 1
                        
                    elif mut_pos < ep : edit_class = 'LHA_edit'; priority = 2 + ep - mut_pos
                    elif mut_pos > ep : edit_class = 'RHA_edit'; priority = 3 + ep + mut_pos
                    
                    # 2/ GC contents가 변화하면 값 증가
                    if gc_fraction(codon) != gc_fraction(mut_codon): priority += 1

                    # 3/ 또 추가할 것이 있나?

                    ###################################################################

                    # Codon 중 intron에 속하는 것은 AA sequence translation에서 제외
                    self.codon_intron_5 = self.cds_start - codon_start
                    codon_intron_3 = codon_end - self.cds_end

                    aa_wt_codon = codon
                    aa_mut_codon = mut_codon
                    
                    if self.codon_intron_5 > 0: 
                        if self.codon_intron_5 % 3 == 0:
                            aa_wt_codon = codon[(self.codon_intron_5 // 3) * 3:]
                            aa_mut_codon = mut_codon[(self.codon_intron_5 // 3) * 3:]
                        
                        else :
                            aa_wt_codon = codon[((self.codon_intron_5 // 3) + 1) * 3:]
                            aa_mut_codon = mut_codon[((self.codon_intron_5 // 3) + 1) * 3:]
                            
                            # partial codon 부분의 mutation filtering
                            partial_len = len(codon) - len(aa_wt_codon)
                            if snv_pos in [partial_len, partial_len-1]: continue 

                    if codon_intron_3 > 0: 
                        if codon_intron_3 % 3 == 0:
                            aa_wt_codon = codon[:-(codon_intron_3 // 3) * 3]
                            aa_mut_codon = mut_codon[:-(codon_intron_3 // 3) * 3]                    
                                                   
                        else:
                            aa_wt_codon = codon[:-((codon_intron_3 // 3) + 1) * 3]
                            aa_mut_codon = mut_codon[:-((codon_intron_3 // 3) + 1) * 3]                                
                        
                            # partial codon 부분의 mutation filtering
                            partial_len = len(aa_wt_codon)
                            if snv_pos in [partial_len, partial_len+1]: continue 
                        
                    aa_wt  = translate(aa_wt_codon)
                    aa_mut = translate(aa_mut_codon)

                    if mut_refpos in self.splicing_adaptor: silent_check = False
                    else                                  : silent_check = aa_wt==aa_mut
                    
                    # synonymous mutation이 cds에 있는지 표기
                    ref_seq = self.ref_seq
                    spacer  = self.spacer
                    cds_start = self.cds_start
                    cds_end   = self.cds_end
                    
                    if strand == '+':
                        abs_mut_pos = ref_seq.find(spacer) + 17 + mut_pos - 1 # 0에서 시작하는 숫자
                    else: 
                        abs_mut_pos = 120 - (reverse_complement(ref_seq).find(spacer) + 17 + mut_pos -1)
                    
                    if (abs_mut_pos >= cds_start) and (abs_mut_pos <= cds_end):
                        mut_in_cds = 'True' # synonymous mutation이 cds에 있는 경우
                    else:
                        mut_in_cds = 'False'  # synonymous mutation이 cds 바깥에 있는 경우   
                    
                    # 전체 결과를 dict에 넣기
                    self.dict_mut['Codon_WT'].append(codon)
                    self.dict_mut['Codon_Mut'].append(mut_codon)
                    self.dict_mut['Codon_MutPos'].append(snv_pos+1)       # First position = 1
                    self.dict_mut['Codon_RefStart'].append(codon_start+1) # First position = 1
                    self.dict_mut['Codon_RefEnd'].append(codon_end+1)     # First position = 1
                    self.dict_mut['RTT_DNA_frame'].append(rtt_frame)
                    self.dict_mut['RTT_DNA_Strand'].append(strand)
                    self.dict_mut['AminoAcid_WT'].append(aa_wt)
                    self.dict_mut['AminoAcid_Mut'].append(aa_mut)
                    self.dict_mut['Silent_check'].append(silent_check)
                    self.dict_mut['Mut_pos'].append(mut_pos)              # First position = 1
                    self.dict_mut['Mut_refpos'].append(mut_refpos)
                    self.dict_mut['Priority'].append(priority) # intended edit (PAM) 위치에 가까울수록 우선
                    self.dict_mut['PAM_Mut'].append(rtt_dna_mut[4:6])
                    self.dict_mut['RTT_DNA_Mut'].append(rtt_dna_mut)
                    self.dict_mut['Edit_class'].append(edit_class)
                    self.dict_mut['Syn_in_CDS'].append(mut_in_cds)

        self.mutations  = pd.DataFrame(self.dict_mut) 
        try: 
            self.synonymous = self.mutations.groupby(by='Silent_check').get_group(True).sort_values(by='Priority').reset_index(drop=True)
            if mut_in_cds_filter:
                self.synonymous = self.synonymous[self.synonymous['Syn_in_CDS'] == 'True']

            return self.synonymous.iloc[0] # Series 형태
        
        except:
            return self.mutations # DataFrame 형태
    # def End: generate

    
    def stack(self, num:int, max_rha_edit:int = 2):
        """만들 수 있는 Synonymous Mut 들의 조합을 추가로 만들어서, 
        그 중에서도 synonymous mut이 존재하는지 확인하고,
        가능한 조합을 return 하는 method.

        Args:
            num (int): Synonymous Mutation을 쌓을 최대 제한 수
            max_rha_edit (int) : 허용할 RHA edit의 최대 제한 수. (지정 안하면 제한 안둠.)

        Returns:
            _type_: str, list
        """        

        # Step1: 중복 자리에 만들어지는 mutation은 제거
        syn_dedup = self.synonymous.drop_duplicates(['Mut_pos']).reset_index(drop=True)
        best_syn  = syn_dedup.iloc[0]
        strand    = best_syn.RTT_DNA_Strand

        # Codon 중 intron에 속하는 것은 AA sequence translation에서 제외
        codon_intron_5 = self.cds_start - best_syn.Codon_RefStart
        codon_intron_3 = best_syn.Codon_RefEnd - self.cds_end

        aa_origin_codon_wt = best_syn.Codon_WT
        aa_origin_codon = best_syn.Codon_Mut

        if codon_intron_5 > 0:
            if codon_intron_5 % 3 == 0:
                aa_origin_codon = best_syn.Codon_Mut[(codon_intron_5 // 3) * 3:]
            else :
                aa_origin_codon = best_syn.Codon_Mut[((codon_intron_5 // 3) + 1) * 3:]
            
            
        if codon_intron_3 > 0: 
            if codon_intron_3 % 3 == 0:
                aa_origin_codon = best_syn.Codon_Mut[:-(codon_intron_3 // 3) * 3]
            else : 
                aa_origin_codon = best_syn.Codon_Mut[:-((codon_intron_3 // 3) + 1) * 3]

        aa_origin = translate(aa_origin_codon)

        selected_mut_codon = best_syn.Codon_Mut

        synMut_cnt = 1
        synMut_RHA_cnt = 1 if best_syn.Edit_class == 'RHA_edit' else 0
        stacked_pos = [best_syn.Mut_pos]

        for i in range(1, len(syn_dedup)):
            
            temp_syn = syn_dedup.iloc[i]
            
            
            if best_syn.Edit_class == 'RHA_edit': 
                print('best_syn is RHA_edit' %self.sID)
                break
            elif best_syn.Edit_class in ['PAM_edit_LHA', 'LHA_edit']:
                if temp_syn.Edit_class == 'PAM_edit_RHA': continue
                elif temp_syn.Edit_class in ['PAM_edit_LHA', 'LHA_edit']:
                    mut_pos = temp_syn.Codon_MutPos - 1
                    mut_nt  = temp_syn.Codon_Mut[mut_pos]

                    stacked_mut_codon = list(selected_mut_codon)
                    stacked_mut_codon[mut_pos] = mut_nt
                    stacked_mut_codon = ''.join(stacked_mut_codon)

                    # Silent Mut check
                    aa_mut_codon = temp_syn.Codon_Mut

                    if codon_intron_5 > 0:
                        if codon_intron_5 % 3 == 0:
                            aa_mut_codon = temp_syn.Codon_Mut[(codon_intron_5 // 3) * 3:]
                        else :
                            aa_mut_codon = temp_syn.Codon_Mut[((codon_intron_5 // 3) + 1) * 3:]
                        
                        
                    if codon_intron_3 > 0: 
                        if codon_intron_3 % 3 == 0:
                            aa_mut_codon = temp_syn.Codon_Mut[:-(codon_intron_3 // 3) * 3]
                        else : 
                            aa_mut_codon = temp_syn.Codon_Mut[:-((codon_intron_3 // 3) + 1) * 3]


                    aa_origin = translate(aa_origin_codon)
                    aa_mut    = translate(aa_mut_codon)

                    if aa_origin == aa_mut:
                        
                        # 만약 조건에 맞는 mutation을 찾으면, origin_codon을 업데이트
                        selected_mut_codon =  stacked_mut_codon
                        aa_origin_codon =  stacked_mut_codon

                        if codon_intron_5 > 0:
                            if codon_intron_5 % 3 == 0:
                                aa_origin_codon = stacked_mut_codon[(codon_intron_5 // 3) * 3:]
                            else :
                                aa_origin_codon = stacked_mut_codon[((codon_intron_5 // 3) + 1) * 3:]
                            
                        if codon_intron_3 > 0: 
                            if codon_intron_3 % 3 == 0:
                                aa_origin_codon = stacked_mut_codon[:-(codon_intron_3 // 3) * 3]
                            else : 
                                aa_origin_codon = stacked_mut_codon[:-((codon_intron_3 // 3) + 1) * 3]
                                
                        synMut_cnt += 1
                        synMut_RHA_cnt += 1 if temp_syn.Edit_class == 'RHA_edit' else 0
                        stacked_pos.append(temp_syn.Mut_pos)

                        if (synMut_cnt >= num) or (synMut_RHA_cnt >= max_rha_edit): break
                else : break
                
            elif best_syn.Edit_class == 'PAM_edit_RHA':
                if temp_syn.Edit_class == 'PAM_edit_RHA': continue
                elif temp_syn.Edit_class in ['PAM_edit_LHA', 'LHA_edit']:
                    mut_pos = temp_syn.Codon_MutPos - 1
                    mut_nt  = temp_syn.Codon_Mut[mut_pos]

                    stacked_mut_codon = list(selected_mut_codon)
                    stacked_mut_codon[mut_pos] = mut_nt
                    stacked_mut_codon = ''.join(stacked_mut_codon)

                    # Silent Mut check
                    aa_mut_codon = temp_syn.Codon_Mut

                    if codon_intron_5 > 0:
                        if codon_intron_5 % 3 == 0:
                            aa_mut_codon = temp_syn.Codon_Mut[(codon_intron_5 // 3) * 3:]
                        else :
                            aa_mut_codon = temp_syn.Codon_Mut[((codon_intron_5 // 3) + 1) * 3:]
                        
                        
                    if codon_intron_3 > 0: 
                        if codon_intron_3 % 3 == 0:
                            aa_mut_codon = temp_syn.Codon_Mut[:-(codon_intron_3 // 3) * 3]
                        else : 
                            aa_mut_codon = temp_syn.Codon_Mut[:-((codon_intron_3 // 3) + 1) * 3]


                    aa_origin = translate(aa_origin_codon)
                    aa_mut    = translate(aa_mut_codon)

                    if aa_origin == aa_mut:
                        
                        # 만약 조건에 맞는 mutation을 찾으면, origin_codon을 업데이트
                        selected_mut_codon =  stacked_mut_codon
                        aa_origin_codon =  stacked_mut_codon

                        if codon_intron_5 > 0:
                            if codon_intron_5 % 3 == 0:
                                aa_origin_codon = stacked_mut_codon[(codon_intron_5 // 3) * 3:]
                            else :
                                aa_origin_codon = stacked_mut_codon[((codon_intron_5 // 3) + 1) * 3:]
                            
                        if codon_intron_3 > 0: 
                            if codon_intron_3 % 3 == 0:
                                aa_origin_codon = stacked_mut_codon[:-(codon_intron_3 // 3) * 3]
                            else : 
                                aa_origin_codon = stacked_mut_codon[:-((codon_intron_3 // 3) + 1) * 3]

                        synMut_cnt += 1
                        synMut_RHA_cnt += 1 if temp_syn.Edit_class == 'RHA_edit' else 0
                        stacked_pos.append(temp_syn.Mut_pos)

                        if (synMut_cnt >= num) or (synMut_RHA_cnt >= max_rha_edit): break
                else : break

        if strand == '+':
            rtt_dna_mut = selected_mut_codon[self.codon_le:len(selected_mut_codon)-self.codon_re]
        else:
            rtt_dna_mut = reverse_complement(selected_mut_codon[self.codon_le:len(selected_mut_codon)-self.codon_re])

        return rtt_dna_mut, stacked_pos
    


def mismatch(seq: str, 
             n: int, 
             start: int = 0, 
             end: int = -1, 
             capital: bool = False,
             full: bool = False,
             ):
    
    '''
    seq  : mismatch를 만들고자 하는 sequence 정보 (DNA 기준, 추후 RNA 추가해주면 좋을듯?)
    n    : mismatch를 만드는 수
    start: target에서 mismatch를 도입할 시작점
    end  : target에서 mismatch를 도입할 종료점
    capital: mismatched nucleotide 표기를 대문자로 할 것인지, True이면 대문자로 표시됨
    full: 모든 mismatched position, WT, Alt, original seq 등 자세한 내용을 DataFrame으로 받을지.
    '''
    
    from itertools import combinations, product
    
    '''
    아직 미완성!!!!
    '''
    
    seq = seq.upper()
    target_seq = seq[start:end]
    
    input_len = len(seq)
    list_seq = list(seq)
    dic = {}
    loc = list(combinations(range(input_len), n))
    nucleo_dic = {"A": ["T","G","C"], 
                  "T": ["A","G","C"], 
                  "G": ["A","T","C"], 
                  "C": ["A","T","G"]}
    
    for i in loc:
        b = list_seq.copy()
        for k in range(len(i)):
            b[i[k]] = nucleo_dic[b[i[k]]]
        lst = list(product(*b))
        for i in lst:
            dic [''.join(i)] = input
            
    return dic
    

class pegRNA:
    '''
    Dev-ing...
    ToDo: RT-PBS combination dict -> DataFrame transformation
    PBS / RTT separated information must included.
    '''
    def __init__(self, wt_seq, ed_seq, edit_type, edit_len,
                pbs_min:int=6, pbs_max:int=17, rtt_max=40,
                pe_system='PE2max'):
        
        self.sWTSeq = wt_seq
        self.sEditedSeq = ed_seq
        self.sAltKey = edit_type + str(edit_len)
        self.sAltType = edit_type
        self.nAltLen = edit_len

        self.nAltIndex = 60
        self.pbs_range = [pbs_min, pbs_max]
        self.rtt_max   = rtt_max
        self.pe_system = pe_system

        self.sGuideKey = ''
        self.sChrID = ''
        self.sStrand = ''
        self.nGenomicPos = 0
        self.nEditIndex = 0
        self.nPBSLen = 0
        self.nRTTLen = 0
        self.sPBSSeq = ''
        self.sRTSeq = ''
        self.sPegRNASeq = ''
        self.list_sSeqs = []
        self.type_sub = 0
        self.type_ins = 0
        self.type_del = 0
        self.dict_sSeqs = {}
        self.dict_sCombos = {}
        self.dict_sOutput = {}

        if   self.sAltType.startswith('sub'): self.type_sub = 1
        elif self.sAltType.startswith('del'): self.type_del = 1
        elif self.sAltType.startswith('ins'): self.type_ins = 1

        
        self.get_sAltNotation()
        self.get_all_RT_PBS(nMinPBS=self.pbs_range[0]-1, nMaxPBS=self.pbs_range[1], nMaxRT=self.rtt_max, pe_system=pe_system)
        self.make_rt_pbs_combinations()

        self.df_out = pd.DataFrame(self.dict_sCombos)

    # def End: get_input


    def show_output(self): return self.df_out

    def get_sAltNotation(self):
        if self.sAltType == 'sub':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[self.nAltIndex:self.nAltIndex + self.nAltLen], self.sEditedSeq[self.nAltIndex:self.nAltIndex + self.nAltLen])

        elif self.sAltType == 'del':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[self.nAltIndex:self.nAltIndex + 1 + self.nAltLen], self.sEditedSeq[self.nAltIndex])

        else:
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[self.nAltIndex], self.sEditedSeq[self.nAltIndex:self.nAltIndex + self.nAltLen + 1])

    # def END: get_sAltNotation

    def get_all_RT_PBS(self, 
                    nMinPBS = 0,
                    nMaxPBS = 17,
                    nMaxRT = 40,
                    nSetPBSLen = 0,
                    nSetRTLen = 0,
                    pe_system = 'PE2'
                    ):
        """
        nMinPBS: If you set specific number, lower than MinPBS will be not generated. Default=0
        nMaxPBS: If you set specific number, higher than MinPBS will be not generated. Default=17
        nMaxRT = : If you set specific number, higher than MinPBS will be not generated. Default=40
        nSetPBSLen = 0  # Fix PBS Len: Set if >0
        nSetRTLen = 0  # Fix RT  Len: Set if >0
        PAM: 4-nt sequence
        """

        nMaxEditPosWin = nMaxRT + 3  # Distance between PAM and mutation

        dict_sWinSize = {'sub': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 2 - 3, 6], 3: [nMaxRT - 3 - 3, 6]},
                        'ins': {1: [nMaxRT - 2 - 3, 6], 2: [nMaxRT - 3 - 3, 6], 3: [nMaxRT - 4 - 3, 6]},
                        'del': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 1 - 3, 6], 3: [nMaxRT - 1 - 3, 6]}}

        
        if 'NRCH' in pe_system: # for NRCH-PE PAM
            dict_sRE = {'+': '[ACGT][ACGT]G[ACGT]|[ACGT][CG]A[ACGT]|[ACGT][AG]CC|[ATCG]ATG', 
                        '-': '[ACGT]C[ACGT][ACGT]|[ACGT]T[CG][ACGT]|G[GT]T[ACGT]|ATT[ACGT]|CAT[ACGT]|GGC[ACGT]|GTA[ACGT]'} 
        else:
            dict_sRE = {'+': '[ACGT]GG[ACGT]', '-': '[ACGT]CC[ACGT]'} # for Original-PE PAM

        for sStrand in ['+', '-']:

            sRE = dict_sRE[sStrand]
            for sReIndex in regex.finditer(sRE, self.sWTSeq, overlapped=True):

                if sStrand == '+':
                    nIndexStart = sReIndex.start()
                    nIndexEnd = sReIndex.end() - 1
                    sPAMSeq = self.sWTSeq[nIndexStart:nIndexEnd]
                    sGuideSeq = self.sWTSeq[nIndexStart - 20:nIndexEnd]
                else:
                    nIndexStart = sReIndex.start() + 1
                    nIndexEnd = sReIndex.end()
                    sPAMSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd])
                    sGuideSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd + 20])

                nAltPosWin = set_alt_position_window(sStrand, self.sAltKey, self.nAltIndex, nIndexStart, nIndexEnd,
                                                    self.nAltLen)

                ## AltPosWin Filter ##
                if nAltPosWin <= 0:             continue
                if nAltPosWin > nMaxEditPosWin: continue

                nPAM_Nick = set_PAM_nicking_pos(sStrand, self.sAltType, self.nAltLen, self.nAltIndex, nIndexStart, nIndexEnd)

                if not check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, self.sAltType, self.nAltLen,
                                        self.nAltIndex): continue

                sPAMKey = '%s,%s,%s,%s,%s,%s,%s' % (
                    self.sAltKey, self.sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq)

                dict_sRT, dict_sPBS = self.determine_PBS_RT_seq(sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen,
                                                        nSetRTLen, nPAM_Nick, nAltPosWin, self.sEditedSeq)

                nCnt1, nCnt2 = len(dict_sRT), len(dict_sPBS)
                if nCnt1 == 0: continue
                if nCnt2 == 0: continue
                
                if sPAMKey not in self.dict_sSeqs:
                    self.dict_sSeqs[sPAMKey] = ''
                self.dict_sSeqs[sPAMKey] = [dict_sRT, dict_sPBS]

            # loop END: sReIndex
        # loop END: sStrand


    # def END: get_all_RT_PBS

    def determine_PBS_RT_seq(self, sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen, nSetRTLen, nPAM_Nick,
                            nAltPosWin, sForTempSeq):
        dict_sPBS = {}
        dict_sRT = {}

        list_nPBSLen = [nNo + 1 for nNo in range(nMinPBS, nMaxPBS)]
        for nPBSLen in list_nPBSLen:

            ## Set PBS Length ##
            if nSetPBSLen:
                if nPBSLen != nSetPBSLen: continue

            if sStrand == '+':
                nPBSStart = nPAM_Nick - nPBSLen  # 5' -> PamNick
                nPBSEnd = nPAM_Nick
                sPBSSeq = sForTempSeq[nPBSStart:nPBSEnd] # sForTempSeq = self.EditedSeq

            else:
                if self.sAltKey.startswith('sub'):
                    nPBSStart = nPAM_Nick
                elif self.sAltKey.startswith('ins'):
                    nPBSStart = nPAM_Nick + self.nAltLen
                elif self.sAltKey.startswith('del'):
                    nPBSStart = nPAM_Nick - self.nAltLen

                sPBSSeq = reverse_complement(sForTempSeq[nPBSStart:nPBSStart + nPBSLen]) # sForTempSeq = self.EditedSeq

            # if END: sStrand

            sKey = len(sPBSSeq)
            if sKey not in dict_sPBS:
                dict_sPBS[sKey] = ''
            dict_sPBS[sKey] = sPBSSeq
        # loop END: nPBSLen

        if sStrand == '+':
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo + 1 for nNo in range(self.nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            elif self.sAltKey.startswith('ins'):
                list_nRTPos = [nNo + 1 for nNo in range(self.nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            else:
                list_nRTPos = [nNo + 1 for nNo in range(self.nAltIndex, (nPAM_Nick + nMaxRT))] ## 수정! ## del2 RHA 3 del1 RHA2
        else:
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 1 - nMaxRT, self.nAltIndex)] ## 수정! ## sub1 sub 3 RHA 0
            else:
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 3 - nMaxRT, self.nAltIndex + self.nAltLen - 1)] ## 수정! ## ins2 최소가 2까지 ins3 RHA 최소 3 #del2 RHA 2 del1 RHA1
        for nRTPos in list_nRTPos:

            if sStrand == '+':
                nRTStart = nPAM_Nick  # PamNick -> 3'
                nRTEnd = nRTPos
                sRTSeq = sForTempSeq[nRTStart:nRTEnd]

            else:
                if self.sAltKey.startswith('sub'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick  # PamNick -> 3'
                elif self.sAltKey.startswith('ins'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick + self.nAltLen  # PamNick -> 3'
                elif self.sAltKey.startswith('del'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick - self.nAltLen  # PamNick -> 3'

                sRTSeq = reverse_complement(sForTempSeq[nRTStart:nRTEnd])

                if not sRTSeq: continue
            # if END: sStrand

            sKey = len(sRTSeq)

            ## Set RT Length ##
            if nSetRTLen:
                if sKey != nSetRTLen: continue

            ## Limit Max RT len ##
            if sKey > nMaxRT: continue

            ## min RT from nick site to mutation ##
            if self.sAltKey.startswith('sub'):
                if sStrand == '+':
                    if sKey < abs(self.nAltIndex - nPAM_Nick): continue
                else:
                    if sKey < abs(self.nAltIndex - nPAM_Nick + self.nAltLen - 1): continue ### 
            else:
                if sStrand == '-':
                    if sKey < abs(self.nAltIndex - nPAM_Nick + self.nAltLen - 1): continue

            if self.sAltKey.startswith('ins'):
                if sKey < nAltPosWin + 1: continue

            if sKey not in dict_sRT:
                dict_sRT[sKey] = ''
            dict_sRT[sKey] = sRTSeq
        # loop END: nRTPos

        return [dict_sRT, dict_sPBS]


    # def END: determine_PBS_RT_seq

    def make_rt_pbs_combinations(self):
        for sPAMKey in self.dict_sSeqs:

            dict_sRT, dict_sPBS = self.dict_sSeqs[sPAMKey]

            list_sRT = [dict_sRT[sKey] for sKey in dict_sRT]
            list_sPBS = [dict_sPBS[sKey] for sKey in dict_sPBS]

            if sPAMKey not in self.dict_sCombos:
                self.dict_sCombos[sPAMKey] = ''
            self.dict_sCombos[sPAMKey] = {'%s,%s' % (sRT, sPBS): {} for sRT in list_sRT for sPBS in list_sPBS}
        # loop END: sPAMKey
    # def END: make_rt_pbs_combinations


def reverse_complement(sSeq):
    dict_sBases = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', 'U': 'U', 'n': '',
                   '.': '.', '*': '*', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a'}
    list_sSeq = list(sSeq)  # Turns the sequence in to a gigantic list
    list_sSeq = [dict_sBases[sBase] for sBase in list_sSeq]
    return ''.join(list_sSeq)[::-1]

# def END: reverse_complement

def set_alt_position_window(sStrand, sAltKey, nAltIndex, nIndexStart, nIndexEnd, nAltLen):
    if sStrand == '+':

        if sAltKey.startswith('sub'):
            return (nAltIndex + 1) - (nIndexStart - 3)
        else:
            return (nAltIndex + 1) - (nIndexStart - 3)

    else:
        if sAltKey.startswith('sub'):
            return nIndexEnd - nAltIndex + 3 - (nAltLen - 1)

        elif sAltKey.startswith('del'):
            return nIndexEnd - nAltIndex + 3 - nAltLen

        else:
            return nIndexEnd - nAltIndex + 3 + nAltLen
        # if END:
    # if END:

# def END: set_alt_position_window


def set_PAM_nicking_pos(sStrand, sAltType, nAltLen, nAltIndex, nIndexStart, nIndexEnd):
    if sStrand == '-':
        nPAM_Nick = nIndexEnd + 3
    else:
        nPAM_Nick = nIndexStart - 3

    return nPAM_Nick

# def END: set_PAM_Nicking_Pos


def check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, sAltType, nAltLen, nAltIndex):
    nUp, nDown = dict_sWinSize[sAltType][nAltLen]

    if sStrand == '+':
        nPAMCheck_min = nAltIndex - nUp + 1
        nPAMCheck_max = nAltIndex + nDown + 1
    else:
        nPAMCheck_min = nAltIndex - nDown + 1
        nPAMCheck_max = nAltIndex + nUp + 1
    # if END:

    if nIndexStart < nPAMCheck_min or nIndexEnd > nPAMCheck_max:
        return 0
    else:
        return 1

# def END: check_PAM_window


'''Model path and list of files
All models from GitHub repository: genet-models
'''

dict_model_info = {
    
    # DeepSpCas9 model
    'SpCas9': {
        # PAM pattern: NGG
        'type': 'DeepSpCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSpCas9',
        'regex': {'+': '[ATGC]{25}GG[ATGC]{3}',
                  '-': '[ATGC]{3}CC[ATGC]{25}',},

        # PAM pattern: NGG + NGA + NAG
        # 'regex': {'+': '[ATGC]{25}G[AG][ATGC]{3}|[ATGC]{25}AG[ATGC]{3}',
        #     '-': '[ATGC]{3}[CT]C[ATGC]{25}|[ATGC]{3}CT[ATGC]{25}',},

    },

    # DeepCas9variants
    'SpCas9-NG': {
        # PAM pattern: NG + NA
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_NG',
        'regex': {'+': '[ATGC]{25}[AG][ATGC]{4}',
                  '-': '[ATGC]{4}[TC][ATGC]{25}',},
    },
    'SpCas9-NRCH': {
        # PAM pattern: NG + NA + NNG
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_NRCH',
        'regex': {'+': '[ATGC]{25}[AG][ATGC]{4}|[ATGC]{26}G[ATGC]{3}',
                  '-': '[ATGC]{4}[TC][ATGC]{25}|[ATGC]{3}C[ATGC]{26}',},
    },
    'SpCas9-NRRH': {
        # PAM pattern: NG + NA
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_NRRH',
        'regex': {'+': '[ATGC]{25}[AG][ATGC]{4}',
                  '-': '[ATGC]{4}[TC][ATGC]{25}',},
    },
    'SpCas9-NRTH': {
        # PAM pattern: NG + NA
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_NRTH',
        'regex': {'+': '[ATGC]{25}[AG][ATGC]{4}',
                  '-': '[ATGC]{4}[TC][ATGC]{25}',},
    },
    'SpCas9-Sc++': {
        # PAM pattern: NNG[CGT]
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_Sc++',
        'regex': {'+': '[ATGC]{26}G[TGC][ATGC]{2}',
                  '-': '[ATGC]{2}[AGC]C[ATGC]{26}',},
    },
    'SpCas9-SpCas9': {
        # PAM pattern: NGG + NGA + NAG
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_SpCas9',
        'regex': {'+': '[ATGC]{25}G[AG][ATGC]{3}|[ATGC]{25}AG[ATGC]{3}',
                  '-': '[ATGC]{3}[CT]C[ATGC]{25}|[ATGC]{3}CT[ATGC]{25}',},
    },
    'SpCas9-SpG': {
        # PAM pattern: NG + NA
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_SpG',
        'regex': {'+': '[ATGC]{25}[AG][ATGC]{4}',
                  '-': '[ATGC]{4}[TC][ATGC]{25}',},
    },
    'SpCas9-SpRY': {
        # PAM pattern: NNNG
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_SpRY',
        'regex': {'+': '[ATGC]{25}[AG][ATGC]{4}|[ATGC]{27}G[ATGC]{2}',
                  '-': '[ATGC]{4}[CT][ATGC]{25}|[ATGC]{2}C[ATGC]{27}',},
    },
    'SpCas9-VRQR': {
        # PAM pattern: NG + NNAG
        'type': 'DeepCas9variants',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepCas9variants/PAM_variant_VRQR',
        'regex': {'+': '[ATGC]{25}G[ATGC]{4}|[ATGC]{26}AG[ATGC]{2}',
                  '-': '[ATGC]{4}C[ATGC]{25}|[ATGC]{2}CT[ATGC]{26}',},
    },


    # DeepSmallCas9
    'CjCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/CjCas9',
        'length': 22,
        't_length': 4+22+8+3,
        'bio_num': 275,
        'model_tag': 'PreTrain-Final-3-3-3-40-40-40-0.001-757-1000-200',
    },
    'efSaCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/efSaCas9',
        'length': 21,
        't_length': 34,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-499-200-50',
    },
    'enCjCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/enCjCas9',
        'length': 22,
        't_length': 37,
        'bio_num': 275,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-403-1000-50',
    },
    'eSaCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/eSaCas9',
        'length': 21,
        't_length': 34,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-484-200-50',
    },
    'Nm1Cas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/Nm1Cas9',
        'length': 23,
        't_length': 4+22+8+3,
        'bio_num': 287,
        'model_tag': 'PreTrain-Final-3-3-3-40-40-40-0.001-972-200-50',
    },
    'Nm2Cas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/Nm2Cas9',
        'length': 23,
        't_length': 4+23+7+3,
        'bio_num': 287,
        'model_tag': 'PreTrain-Final-3-3-3-40-40-40-0.001-267-200-200',
    },
    'SaCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SaCas9',
        'length': 21,
        't_length': 4+21+6+3,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-20-20-20-0.001-977-200-50',
    },
    'SaCas9-HF': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SaCas9-HF',
        'length': 21,
        't_length': 34,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-496-200-50',
    },
    'SaCas9-KKH': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SaCas9-KKH',
        'length': 21,
        't_length': 4+21+6+3,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-40-40-40-0.001-997-1000-50',
    },
    'SaCas9-KKH-HF': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SaCas9-KKH-HF',
        'length': 21,
        't_length': 34,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-482-200-50',
    },
    'Sa-SlugCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/Sa-SlugCas9',
        'length': 21,
        't_length': 32,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-493-1000-50',
    },
    'SauriCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SauriCas9',
        'length': 21,
        't_length': 4+21+4+3,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-40-40-40-0.001-940-1000-50',
    },
    'SauriCas9-KKH': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SauriCas9-KKH',
        'length': 21,
        't_length': 4+21+4+3,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-20-20-20-0.001-717-200-50',
    },
    'SlugCas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SlugCas9',
        'length': 21,
        't_length': 32,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-476-200-50',
    },
    'SlugCas9-HF': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/SlugCas9-HF',
        'length': 21,
        't_length': 32,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-426-1000-50',
    },
    'sRGN': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/sRGN3.1',
        'length': 21,
        't_length': 32,
        'bio_num': 263,
        'model_tag': 'PreTrain-Final-3-3-3-10-10-10-0.001-497-1000-50',
    },
    'St1Cas9': {
        'type': 'DeepSmallCas9',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepSmallCas9/St1Cas9',
        'length': 19,
        't_length': 4+19+6+3,
        'bio_num': 239,
        'model_tag': 'PreTrain-Final-3-3-3-40-40-40-0.001-639-1000-50',
    },

    # DeepPrime / DeepPrime-FT models
    'PE2-HEK293T': {
        'type': 'DeepPrime_base',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_base'
    },
    'PE2-HCT116': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_HCT116_PE2_Opti_220428'
    },
    'PE2-MDA-MB-231': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_MDA_PE2_Opti_220428'
    },
    'PE2max-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_PE2max_Opti_220428'
    },
    'PE2max-A549': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_A549_PE2max_Opti_221114'
    },
    'PE2max-DLD1': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_DLD1_PE2max_Opti_221114'
    },
    'PE2max-HeLa': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_HeLa_PE2max_Opti_220815'
    },
    'PE2max-e-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_PE2max_epegRNA_Opti_220428'
    },
    'PE2max-e-A549': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_A549_PE2max_epegRNA_Opti_220428'
    },
    'PE4max-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_PE4max_Opti_220728'
    },
    'PE4max-A549': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_A549_PE4max_Opti_220728'
    },
    'PE4max-DLD1': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_DLD1_PE4max_Opti_220728'
    },
    'PE4max-e-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_PE4max_epegRNA_Opti_220428'
    },
    'PE4max-e-A549': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_A549_PE4max_epegRNA_Opti_220428'
    },

    'PE7max-e-A549': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_FT_A549_PE7_epegRNA_dataset'
    },
    'PE7max-A549': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_FT_A549_PE7_pegRNA_dataset'
    },

    'PE7max-e-DLD1': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_FT_DLD1_PE7_epegRNA_dataset'
    },
    'PE7max-DLD1': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_FT_DLD1_PE7_pegRNA_dataset'
    },

    'PE7max-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_FT_HEK293T_PE7_pegRNA_dataset_240909'
    },

    'NRCH_PE2-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_NRCH_PE2_Opti_220428'
    },
    'NRCH_PE2max-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_NRCH-PE2max_Opti_220815'
    },
    'NRCH_PE4max-DLD1': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_DLD1_NRCHPE4max_Opti_220728'
    },
    'NRCH_PE4max-NIH3T3': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_NIH_NRCHPE4max_Opti_220815'
    },
    'PE2-Off-HEK293T': {
        'type': 'DeepPrime-Off',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DeepPrime_off'
    },
    'sRGN-PE2max-HEK293T': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_293T_sRGN-PE2max'
    },
    'sRGN-PE2max-DLD1': {
        'type': 'DeepPrime-FT',
        'repo': 'Goosang-Yu/genet-models/main/genet_models',
        'path': 'DeepPrime/DP_variant_DLD1_sRGN-PE2max'
    },
}


dict_model_requests = {
    'DeepSpCas9': [
        '__init__.py',
        'PreTrain-Final-3-5-7-100-70-40-0.001-550-80-60.data-00000-of-00001', 
        'PreTrain-Final-3-5-7-100-70-40-0.001-550-80-60.index',
        'PreTrain-Final-3-5-7-100-70-40-0.001-550-80-60.meta',
    ],
    
    'DeepCas9variants': [
        '__init__.py',
        'DeepCas9variants_model_WeightQuantization.tflite',
    ],

    'DeepPrime_base': [
        '__init__.py',
        'model_0.pt',
        'model_1.pt',
        'model_2.pt',
        'model_3.pt',
        'model_4.pt',
        'mean_231124.csv',
        'std_231124.csv',
    ],

    'DeepPrime-FT': [
        '__init__.py',
        'final_model_0.pt',
        'final_model_1.pt',
        'final_model_2.pt',
        'final_model_3.pt',
        'final_model_4.pt',
        'final_model_5.pt',
        'final_model_6.pt',
        'final_model_7.pt',
        'final_model_8.pt',
        'final_model_9.pt',
        'final_model_10.pt',
        'final_model_11.pt',
        'final_model_12.pt',
        'final_model_13.pt',
        'final_model_14.pt',
        'final_model_15.pt',
        'final_model_16.pt',
        'final_model_17.pt',
        'final_model_18.pt',
        'final_model_19.pt',
        'mean_231124.csv',
        'std_231124.csv',
    ],

    'DeepPrime-Off': [
        '__init__.py',
        'final_model_0.pt',
        'final_model_1.pt',
        'final_model_2.pt',
        'final_model_3.pt',
        'final_model_4.pt',
        'mean_231124.csv',
        'std_231124.csv',
    ],
}





from Bio.Seq import reverse_complement



def twinPE(spacer_f:str, spacer_r:str, replace_seq:str, overlap:int=25, pbs_f:int=12, pbs_r:int=12) -> str:
    """TwinPE에 필요한 pegRNA 2개의 RT-PBS를 디자인해주는 함수

    Args:
        spacer_f (str): pegRNA upper strand의 spacer
        spacer_r (str): pegRNA bottom strand의 spacer
        replace_seq (str): TwinPE로 replacement 하고 싶은 sequence
        overlap (int, optional): TwinPE로 만들어지는 cDNA들의 overlap 길이. Defaults to 25.
        pbs_f (int, optional): _description_. Defaults to 12.
        pbs_r (int, optional): _description_. Defaults to 12.

    Returns:
        str: RT-PBS 2개 세트를 만들어 줌
    """

    if len(replace_seq) < overlap:
        raise ValueError('The length of replace sequence must be longer than overlap length.')

    spacer_f    = spacer_f.upper()
    spacer_r    = spacer_r.upper()
    replace_seq = replace_seq.upper()

    pbs_f = reverse_complement(spacer_f[17-pbs_f:17])
    pbs_r = reverse_complement(spacer_r[17-pbs_r:17])

    RepSeq_center = int(len(replace_seq)/2)
    half_overlap  = int(overlap/2)

    cdna_f = reverse_complement(replace_seq[:RepSeq_center+half_overlap])
    cdna_r = replace_seq[RepSeq_center-(overlap-half_overlap):]

    rtpbs_1 = cdna_f + pbs_f
    rtpbs_2 = cdna_r + pbs_r

    return rtpbs_1, rtpbs_2




import os, gzip
from Bio import SeqIO
from genet.database import GetGenome, DFConverter



class Gene:
    def __init__(self, gene:str, seq_type:str='CDS', variant:str='MANE Select', spacies:str='Homo sapiens', gbk:str=None, gff:str=None) -> None:
        """GenBank file에서 feature에 있는 정보를 기반으로 원하는 유전자에 대한 정보를 찾고 파싱하는 함수. 

        Args:
            gene (str): 찾고 싶은 유전자의 이름
            seq_type (str, optional): 유전자의 정보 타입 (exon, CDS, transcript 등). Defaults to 'CDS'.
            variant (str, optional): 만약 transcript variants를 특정 id로 지정하고 싶다면 설정. Defaults to 'MANE Select'.

            spacies (str, optional): Reference genome을 가져올 종의 이름. Defaults to 'Homo sapiens'.
            gbk (str, optional): GenBank file의 위치 경로. Defaults (None)로 지정되어 있다면, 현재 경로에 spacies 이름으로 폴더가 만들어지고, 여기에 파일이 자동으로 다운로드 된다. 
            gff (str, optional): Gene feature file의 위치 경로. Defaults (None)로 지정되어 있다면, 현재 경로에 spacies 이름으로 폴더가 만들어지고, 여기에 파일이 자동으로 다운로드 된다. 
        """

        converter = DFConverter()

        print('Load GenBank file')

        if gbk == None: 
            genome = GetGenome(spacies)
            content = genome.contents()
            
            for f in content:
                if f.endswith('_genomic.gbff.gz'): 
                    gbk_path = f'{spacies.replace(' ', '_')}/{f}'
                    if not os.path.exists(gbk_path): genome.download(f, spacies.replace(' ', '_'))

                    gbk_handle  = SeqIO.parse(gzip.open(gbk_path, "rt"), "genbank")

        else: 
            gbk_handle  = SeqIO.parse(gzip.open(gbk, "rt"), "genbank")
                    
        print('Load Feature file')

        if gff == None: 
            genome = GetGenome(spacies)
            content = genome.contents()
            
            for f in content:
                if f.endswith('_genomic.gff.gz'): 
                    gff_path = f'{spacies.replace(' ', '_')}/{f}'
                    if not os.path.exists(gff_path): genome.download(f, spacies.replace(' ', '_'))

                    df_gff = converter.convert(gff_path)

        else: 
            df_gff = converter.convert(gff)


        pass



    

    


import pandas as pd


class MAGeCKAnalyzer:
    def __init__(self, summary_file:str, anotation:dict=None):
        """MAGeCK test file (summary file)을 기반으로 DataFrame 형태로 변환해주면서
        다양한 분석에 응용할 수 있는 객체

        Args:
            summary_file (str): MAGeCK test output (summary) file.
            anotation (dict, optional): _description_. Defaults to None.
        """        
        self.data   = pd.read_csv(summary_file, sep='\t').set_index('id')
        self.simple = self._make_df_simple(self.data)

    def _get_pvalue(self):
        '''MAGeCK test 결과 파일에서 p-value를 가져오는 method.
        rlqhswjrdmfh pos / neg score 중에서 작은 것을 가져온다.
        ToDo: p-value / FDR / 등 다른 것들에 대해서 가져오는 것도 찾아보기?
        '''
        pos  = self.data['pos|score']
        neg  = self.data['neg|score']
        pval = [min(pos.loc[i], neg.loc[i]) for i in self.data.index]

        return pval
    
    def _make_df_simple(self, data):
        '''MAGeCK test 결과 파일에서 핵심 값들만 가져온 것 (LFC / p-value).'''

        df_simple = data[['pos|lfc']].copy()
        df_simple['p-value'] = self._get_pvalue()
        df_simple.columns = ['LFC', 'p-value']

        return df_simple

    def volcano(self):
        '''Make volcano plot'''


    def kde(self):
        '''Make KDE joyplot'''


    def positive(self, lfc_cut:float, pvalue_cut:float):
        '''get positive hits depending LFC / p-values'''

    def negative(self, lfc_cut:float, pvalue_cut:float):
        '''get negative hits depending LFC / p-values'''




class MAGeCK:
    def __init__(self, ):
        print('MAGeCK object')




    def count(self,):
        '''MAGeCK count method'''


        


import pandas as pd
import numpy as np
import editdistance

from Bio import SeqIO
from tqdm import tqdm




def make_df_umi(list_barcode:list, data_path:str, len_umi:int) -> pd.DataFrame:
    """ A function that separates UMIs by barcode in NGS read files 
    and creates a DataFrame summarizing the read counts.
    ---
    
    ### Args:
        list_barcode (list): List containing barcodes. pd.Series also acceptable.
        data_path (str): The path of NGS data file. FASTQ or FASTA file can be used.
        len_umi (int): The length of UMI for counting.

    ### Raises:
        ValueError: NGS data format or path error. 
        ValueError: The lengths of barcode error. Barcode length should be identical.
        ValueError: No barcode error. Check your barcode list.

    ### Returns:
        _type_: pd.DataFrame
    """    
     
    # Input checker: Check and Determine data file format
    if   data_path.split('.')[-1] in ['fastq', 'fq']: data_format = 'fastq'
    elif data_path.split('.')[-1] in ['fasta', 'fa']: data_format = 'fasta'
    else: raise ValueError('Please check your input: data_path')

    # Input checker: Check barcode length. The length of barcodes should be identical.
    list_bc_len = [len(bc) for bc in list_barcode]
    if np.std(list_bc_len) != 0: raise ValueError('Please check your input: The lengths of barcode is not identical')
    if len(list_barcode)   == 0: raise ValueError('Please check your input: No barcde found in list_barcode')
    len_bc = list_bc_len[0]
    

    # Step1: Make dictionary containing Barcodes and founded UMIs
    dict_bc = {}
    for bc in list_barcode: dict_bc[bc] = {}

    list_seq = [str(s.seq) for s in SeqIO.parse(data_path, data_format)]

    for _seq in tqdm(list_seq,
                total = len(list_seq),
                desc = 'Barcode/UMI sorting',
                ncols = 70,
                ascii = ' =',
                leave = True
                ):
        
        _bc  = _seq[:len_bc]
        _umi = _seq[len_umi:]
        
        if _bc in dict_bc: 
            if _umi in dict_bc[_bc]: dict_bc[_bc][_umi] += 1
            else                   : dict_bc[_bc][_umi] = 1
        
        else: continue

    # Step2: Make DataFrame as output
    list_df_temp = []

    for bc in tqdm(dict_bc,
                total = len(dict_bc),
                desc = 'Make output ',
                ncols = 70,
                ascii = ' =',
                leave = True
                ):
        
        list_bc  = []
        list_umi = []
        list_cnt = []
        
        for umi in dict_bc[bc]:
            list_bc.append(bc)
            list_umi.append(umi)
            list_cnt.append(dict_bc[bc][umi])
            
        df_temp = pd.DataFrame()
        df_temp['Barcode'] = list_bc
        df_temp['UMI']     = list_umi
        df_temp['count']   = list_cnt
        
        list_df_temp.append(df_temp)
        
        
    df_out = pd.concat(list_df_temp).reset_index(drop=True)

    return df_out


def count_mismatch(ref_seq:str, match_seq:str) -> int:
    ref_seq        = list(ref_seq)
    match_seq      = list(match_seq)
    mismatch_count = 0
    
    for ref, m_seq in zip(ref_seq, match_seq):
        if ref != m_seq: mismatch_count += 1
    
    return mismatch_count

def custom_edit_distance(a:str, b:str):
    dp = [[0] * (len(b)+1) for _ in range(len(a) + 1)]
    
    for i in range(1, len(a)+1): dp[i][0] = i
    for j in range(1, len(b)+1): dp[0][j] = j

    for i in range(1, len(a)+1):
        for j in range(1, len(b)+1):
            if a[i-1] == b[j-1]: dp[i][j] = dp[i-1][j-1]
            else               : dp[i][j] = min(dp[i-1][j-1], dp[i-1][j], dp[i][j-1]) + 1

    return dp[-1][-1]


def collapse_umi(df_umi:pd.DataFrame, threshold:int=1, col_bc:str=None, col_umi:str=None, col_count:str=None,):
    """Combine different UMIs when they varied by only less or same nucleotide threshold. 
    Input dataframe should contain barcode, UMI, and count informations.
    
    Args:
        df_umi (pd.DataFrame): Barcode, UMI and count information. 
        threshold (int, optional): UMI collapse threshold. Defaults to 1.
        col_bc (str, optional): If Barcode is not in first column of df_umi, select the column name of barcode. Defaults to None.
        col_umi (str, optional): If UMI is not in first column of df_umi, select the column name of UMI. Defaults to None.
        col_count (str, optional): If UMI counts is not in first column of df_umi, select the column name of UMI counts. Defaults to None.

    Raises:
        ValueError: For col_bc, col_umi, and col_count. Not found error.

    Returns:
        _type_: pd.DataFrame
    """    
    list_col_names = list(df_umi.columns)

    # Default Barcode column index location = 0 (first column)
    if col_bc    == None: col_bc    = list_col_names[0]
    if col_umi   == None: col_umi   = list_col_names[1]
    if col_count == None: col_count = list_col_names[2]
    
    try   : bc_group = df_umi.groupby(by=[col_bc])
    except: raise ValueError(f'Can not find Barcode colume - {col_bc}. Please check input')
    
    try   : list_umi = df_umi[col_umi]
    except: raise ValueError(f'Can not find Barcode colume - {col_umi}. Please check input')
    
    list_bc = list(df_umi[col_bc].unique())
    list_df = []
    dict_collaped = {'Barcode':[], 'UMI':[], 'count'  :[]}
    
    for bc in tqdm(list_bc,
            total = len(list_bc),       ## 전체 진행수
            desc = 'Barcode collapsed', ## 진행률 앞쪽 출력 문장
            ncols = 70,                 ## 진행률 출력 폭 조절
            ascii = ' =',               ## 바 모양, 첫 번째 문자는 공백이어야 작동
            leave = True
            ):
    # for bc in list_bc:
        df_bc = bc_group.get_group(bc)
        
        list_umi   = list(df_bc[col_umi])
        list_cnt   = list(df_bc[col_count])
        list_check = []

        for i, umi in enumerate(list_umi):
            if umi in list_check: continue
            dict_collaped[col_bc].append(bc)
            dict_collaped[col_umi].append(umi)
            dict_collaped[col_count].append(list_cnt[i])
                
            list_residual_umi = list_umi[i+1:]
            
            for _i_res, _res_umi in enumerate(list_residual_umi):
                if _res_umi in list_check: continue
                if count_mismatch(umi, _res_umi) <= threshold:
                    dict_collaped['count'][-1] += list_cnt[i+_i_res+1]
                    list_check.append(_res_umi)
                
        # list_df.append(pd.DataFrame.from_dict(data=dict_collaped, orient='columns'))
    
    print('Making final DataFrame')
    df_out = pd.DataFrame.from_dict(data=dict_collaped, orient='columns')
        
    
    return df_out


# genet package and modules 
from genet.predict.PredUtils import *
from genet.predict.Nuclease import SpCas9
from genet.models import LoadModel
from genet.database import GetGenome, GetChromosome

# python standard packages
import os, sys, regex, gzip
import numpy as np
import pandas as pd
from glob import glob
from tqdm import tqdm
from time import time

# pytorch package and modules 
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

# biopython package and modules 
from Bio import SeqIO
from Bio.Seq import Seq, transcribe, back_transcribe, reverse_complement, reverse_complement_rna
from Bio.SeqUtils import MeltingTemp as mt
from Bio.SeqUtils import gc_fraction as gc

from RNA import fold_compound

np.set_printoptions(threshold=sys.maxsize)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class DeepPrime:

    def __init__(self, sequence: str, name:str='SampleName', pam:str = 'NGG',
                 pbs_min:int = 7, pbs_max:int = 15,
                 rtt_min:int = 0, rtt_max:int = 40, 
                 spacer_len:int=20,
                ):
        """DeepPrime: pegRNA activity prediction models

        Args:
            sequence (str): Sequence to prime editing. Intended prime editing should marked with parenthesis.
            name (str, optional): Sample ID for pegRNAs. Defaults to 'SampleName'
            pam (str, optional): PAM sequence. Available PAMs are NGG, NGA, NAG, NRCH. Defaults to 'NGG'.
            pbs_min (int, optional): Minimum length of PBS (1-17). Defaults to 7.
            pbs_max (int, optional): Maximum length of PBS (1-17). Defaults to 15.
            rtt_min (int, optional): Minimum length of RTT (0-40). Defaults to 0.
            rtt_max (int, optional): Maximum length of RTT (0-40). Defaults to 40.
        """        
        
        # input parameters
        self.input_params = {
            'name'      : name,
            'sequence'  : sequence,
            'nAltIndex' : 60,
            'spacerlen' : spacer_len,
            'pam'       : pam,
            'pbs_min'   : pbs_min,
            'pbs_max'   : pbs_max,
            'rtt_min'   : rtt_min,
            'rtt_max'   : rtt_max,
            'wt_seq'    : None,
            'ed_seq'    : None,
            'edit_type' : None,
            'edit_len'  : None,
        }

        # Check input parameters
        self.dp_input = DeepPrimeInputProcessor(self.input_params)
        self.input_params = self.dp_input.check_input()

        ## DeepPrime Feature Extraction 
        cFeat = PEFeatureExtraction(self.input_params)

        cFeat.input_id = name
        cFeat.get_all_RT_PBS(self.input_params)
        cFeat.make_rt_pbs_combinations()
        cFeat.determine_seqs()
        cFeat.determine_secondary_structure()
        self.features = cFeat.make_output_df()
        
        del cFeat

        if len(self.features) > 0:
            self.list_Guide30 = [WT74[:30] for WT74 in self.features['Target']]
            self.features['DeepSpCas9_score'] = SpCas9().predict(self.list_Guide30)['SpCas9']
            self.pegRNAcnt = len(self.features)
        
        else:
            print('\nsID:', name)
            print('DeepPrime only support RTT length upto 40nt')
            print('There are no available pegRNAs, please check your input sequences.\n')
            self.pegRNAcnt = 0

    # def __init__: END


    def predict(self, pe_system:str, cell_type:str = 'HEK293T', show_features:bool = False, report=False) -> pd.DataFrame:
        """_summary_
    
        Args:
            pe_system (str): Available PE systems are PE2, PE2max, PE4max, NRCH_PE2, NRCH_PE2max, NRCH_PE4max
            cell_type (str, optional): Available Cell types are HEK293T, HCT116, MDA-MB-231, HeLa, DLD1, A549, NIH3T3. Defaults to 'HEK293T'.
            show_features (bool, optional): _description_. Defaults to False.
            report (bool, optional): _description_. Defaults to False.

        Returns:
            pd.DataFrame: 각 pegRNA와 target쌍 마다의 DeepPrime prediction score를 계산한 결과를 DataFrame으로 반환.
        """
        
        # Load models
        model_info = LoadModel('DeepPrime', pe_system, cell_type)
        model_dir  = model_info.model_dir

        # Check pe_system is available for PAM
        self.dp_input.check_pe_type(pe_system)

        # Data preprocessing for deep learning model
        df_all = self.features.copy()

        os.environ['CUDA_VISIBLE_DEVICES']='0'
        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        mean = pd.read_csv(f'{model_dir}/mean_231124.csv', header=None, index_col=0).squeeze()
        std  = pd.read_csv(f'{model_dir}/std_231124.csv',  header=None, index_col=0).squeeze()

        test_features = select_cols(df_all)

        g_test = seq_concat(df_all)
        x_test = (test_features - mean) / std

        g_test = torch.tensor(g_test, dtype=torch.float32, device=device)
        x_test = torch.tensor(x_test.to_numpy(), dtype=torch.float32, device=device)

        models = [m_files for m_files in glob(f'{model_dir}/*.pt')]
        preds  = []

        for m in models:
            model = GeneInteractionModel(hidden_size=128, num_layers=1).to(device)
            model.load_state_dict(torch.load(m, map_location=device))
            model.eval()
            with torch.no_grad():
                g, x = g_test, x_test
                g = g.permute((0, 3, 1, 2))
                pred = model(g, x).detach().cpu().numpy()
            preds.append(pred)
        
        # AVERAGE PREDICTIONS
        preds = np.squeeze(np.array(preds))
        preds = np.mean(preds, axis=0)
        preds = np.exp(preds) - 1

        df_all.insert(1, f'{pe_system}_score', preds)


        if   show_features == False: return df_all.iloc[:, :11]
        elif show_features == True : return df_all

    # def predict: END


class DeepPrimeInputProcessor:
    def __init__(self, input_params:dict):
        """Make DeepPrime inputs from input sequence.
        
        Args:
            sequence (str): Sequence to prime editing. Intended prime editing should marked with parenthesis.

        Raises:
            ValueError: _description_

        Returns:
            dict: DeepPrime inputs from input sequence
        """    

        self.input_params = input_params

        sequence = self.input_params['sequence']

        f_context, res_seq  = sequence.split('(')
        edit_seq, r_context = res_seq.split(')')
        wt, ed = edit_seq.split('/')

        if len(wt) == len(ed): edit_type = 'sub'; edit_len = len(ed)
        elif wt == '': edit_type = 'ins'; edit_len = len(ed)
        elif ed == '': edit_type = 'del'; edit_len = len(wt)
        else: raise ValueError('Not supported edit type.')

        self.input_params.update({
            'wt_seq': f_context[-60:] + (wt + r_context)[:61],
            'ed_seq' : f_context[-60:] + (ed + r_context)[:61],
            'edit_type': edit_type,
            'edit_len': edit_len,
        })
        

    def check_input(self):
        
        if len(self.input_params['wt_seq']) != 121:
            raise ValueError('Please check your input sequence. The length of context sequence is not enough.')
        
        if len(self.input_params['ed_seq']) != 121:
            raise ValueError('Please check your input sequence. The length of context sequence is not enough.')

        if self.input_params['pbs_min'] < 1:
            raise ValueError('Please check your input: pbs_min. Please set PBS max length at least 1nt')
        
        if self.input_params['pbs_max'] > 17:
            raise ValueError('Please check your input: pbs_max. Please set PBS max length upto 17nt')
        
        if self.input_params['rtt_max'] > 40:
            raise ValueError('Please check your input: rtt_max. Please set RTT max length upto 40nt')

        if self.input_params['edit_type'] not in ['sub', 'ins', 'del']:
            raise ValueError('Please check your input: edit_type. Available edit style: sub, ins, del')
        
        if self.input_params['pam'] not in ['NGG', 'NRCH', 'NAG', 'NGA', 'NNGG']:
            raise ValueError('Please check your input: pam. Available PAM: NGG, NGA, NAG, NRCH, NNGG')

        if self.input_params['edit_len'] > 3:
            raise ValueError('Please check your input: edit_len. Please set edit length upto 3nt. Available edit length range: 1~3nt')
        
        if self.input_params['edit_len'] < 1:
            raise ValueError('Please check your input: edit_len. Please set edit length at least 1nt. Available edit length range: 1~3nt')

        return self.input_params
    
    # def check_input: END

    def check_pe_type(self, pe_system:str):

        pam = self.input_params['pam']
        
        if pam == 'NRCH' and pe_system not in ['NRCH_PE2', 'NRCH_PE2max', 'NRCH_PE4max']:
            raise ValueError(f'{pam} PAM is not available for {pe_system}. Please check PAM or pe_system. ')
        
        if pam == 'NNGG' and pe_system not in ['sRGN-PE2max']:
            raise ValueError(f'{pam} PAM is not available for {pe_system}. Please check PAM or pe_system. ')

        return None
    
    # def check_input: END


class DeepPrimeGuideRNA:

    def __init__(self, sID:str, target:str, pbs:str, rtt:str, 
                 edit_len:int, edit_pos:int, edit_type:str, 
                 spacer_len:int=20 
                 ):
        """A pipeline used when running DeepPrime on a pre-designed pegRNA.

        Args:
            sID (str): ID of the pegRNA.
            target (str): Target sequence, fixed length of 74nt.
            pbs (str): PBS sequence.
            rtt (str): RTT sequence.
            edit_len (int): Length of prime editing. Available edit length range: 1-3nt.
            edit_pos (int): Position of prime editing. Available edit position range: 1-40nt.
            edit_type (str): Type of prime editing. Available edit style: sub, ins, del.

        Raises:
            ValueError: Raised if the target sequence length is not 74nt.
            ValueError: Raised if the edit length is not 1, 2, or 3.
            ValueError: Raised if the edit type is not one of sub, ins, del.
            
        ### Examples:
        ``` python
        from genet.predict import DeepPrimeGuideRNA

        target    = 'ATAAAAGACAACACCCTTGCCTTGTGGAGTTTTCAAAGCTCCCAGAAACTGAGAAGAACTATAACCTGCAAATG'
        pbs       = 'GGCAAGGGTGT'
        rtt       = 'CGTCTCAGTTTCTGGGAGCTTTGAAAACTCCACAA'
        edit_len  = 1
        edit_pos  = 34
        edit_type = 'sub'

        pegrna = DeepPrimeGuideRNA('pegRNA_test', target=target, pbs=pbs, rtt=rtt,
                                edit_len=edit_len, edit_pos=edit_pos, edit_type=edit_type)

        pe2max_score = pegrna.predict('PE2max')
        ```
        """        

        # PBS와 RTT는 target 기준으로 reverse complementary 방향으로 있어야 함.
        # PBS와 RTT를 DNA/RNA 중 어떤 것으로 input을 받아도, 전부 DNA로 변환해주기.

        target = target.upper()
        pbs    = back_transcribe(pbs).upper()
        rtt    = back_transcribe(rtt).upper()

        if len(target) != 74: raise ValueError('Please check your input: target. The length of target should be 74nt')
        if reverse_complement(pbs) != target[21-len(pbs):21]: raise ValueError('Please check your input: target, pbs sequence and position.')
        if edit_len not in [1, 2, 3]: raise ValueError('Please check your input: edit_len. The length of edit should be 1, 2, or 3')

        self.spacer = target[24-spacer_len:24]
        self.rtpbs  = rtt + pbs

        # Check edit_type input and determine type dependent features
        if   edit_type == 'sub': type_sub=1; type_ins=0; type_del=0; rha_len=len(rtt)-edit_pos-edit_len+1
        elif edit_type == 'ins': type_sub=0; type_ins=1; type_del=0; rha_len=len(rtt)-edit_pos-edit_len+1
        elif edit_type == 'del': type_sub=0; type_ins=0; type_del=1; rha_len=len(rtt)-edit_pos+1
        else: raise ValueError('Please check your input: edit_type. Available edit style: sub, ins, del')


        # pegRNA Tm feature
        seq_Tm1    = transcribe(pbs)
        seq_Tm2    = target[21:21+len(rtt)]

        if edit_type == 'sub':
            seq_Tm3 = target[21:21 + len(rtt)]
            sTm4antiSeq = reverse_complement(target[21:21 + len(rtt)])
        elif edit_type == 'ins':
            seq_Tm3 = target[21:21 + len(rtt) - edit_len]
            sTm4antiSeq = reverse_complement(target[21:21 + len(rtt) - edit_len])
        elif edit_type == 'del':
            seq_Tm3 = target[21:21 + len(rtt) + edit_len]
            sTm4antiSeq = reverse_complement(target[21:21 + len(rtt) + edit_len])                    
        
        seq_Tm4 = [back_transcribe(reverse_complement(rtt)), sTm4antiSeq] # 원래 코드에는 [sRTSeq, sTm3antiSeq]

        seq_Tm5 = transcribe(rtt) # 원래 코드: reverse_complement(sRTSeq.replace('A', 'U'))

        fTm1 = mt.Tm_NN(seq=Seq(seq_Tm1), nn_table=mt.R_DNA_NN1)
        fTm2 = mt.Tm_NN(seq=Seq(seq_Tm2), nn_table=mt.DNA_NN3)
        fTm3 = mt.Tm_NN(seq=Seq(seq_Tm3), nn_table=mt.DNA_NN3)
        

        # 이 부분이 사실 의도된 feature는 아니긴 한데... 이미 이렇게 모델이 만들어졌음...
        for sSeq1, sSeq2 in zip(seq_Tm4[0], seq_Tm4[1]):
            try:
                fTm4 = mt.Tm_NN(seq=sSeq1, c_seq=sSeq2, nn_table=mt.DNA_NN3)
            except ValueError:
                fTm4 = 0

        ######### 이 부분이 문제 ###################################################
        # 이미 DeepPrime이 이 형태로 학습되었으니... 그대로 사용.
        
        fTm5 = mt.Tm_NN(seq=Seq(seq_Tm5), nn_table=mt.R_DNA_NN1)

        ############################################################################

        # MFE_3 - RT + PBS + PolyT
        seq_MFE3 = self.rtpbs + 'TTTTTT'
        sDBSeq, fMFE3 = fold_compound(seq_MFE3).mfe()

        # MFE_4 - spacer only
        seq_MFE4 = 'G' + self.spacer[1:]
        sDBSeq, fMFE4 = fold_compound(seq_MFE4).mfe()


        self.dict_feat = {
            # Enter your sample's ID
            'ID'                         : [sID],

            # pegRNA sequence information
            'Spacer'                     : [self.spacer],
            'RT-PBS'                     : [self.rtpbs],
            
            # pegRNA length feature
            'PBS_len'                    : [len(pbs)],
            'RTT_len'                    : [len(rtt)],
            'RT-PBS_len'                 : [len(self.rtpbs)],
            'Edit_pos'                   : [edit_pos],
            'Edit_len'                   : [edit_len],
            'RHA_len'                    : [rha_len],

            # Target sequence information
            'Target'                     : [target],
            'Masked_EditSeq'             : ['x'*(21-len(pbs)) + reverse_complement(self.rtpbs) + 'x'*(74-21-len(rtt))],

            # Edit type information
            'type_sub'                   : [type_sub],
            'type_ins'                   : [type_ins],
            'type_del'                   : [type_del],

            # pegRNA Tm feature
            'Tm1_PBS'                    : [fTm1],
            'Tm2_RTT_cTarget_sameLength' : [fTm2],
            'Tm3_RTT_cTarget_replaced'   : [fTm3], 
            'Tm4_cDNA_PAM-oppositeTarget': [fTm4],
            'Tm5_RTT_cDNA'               : [fTm5],
            'deltaTm_Tm4-Tm2'            : [fTm4 - fTm2],

            # pegRNA GC feature
            'GC_count_PBS'               : [pbs.count('G') + pbs.count('C')],
            'GC_count_RTT'               : [rtt.count('G') + rtt.count('C')],
            'GC_count_RT-PBS'            : [self.rtpbs.count('G') + self.rtpbs.count('C')],
            'GC_contents_PBS'            : [100 * gc(pbs)],
            'GC_contents_RTT'            : [100 * gc(rtt)],
            'GC_contents_RT-PBS'         : [100 * gc(self.rtpbs)],

            # pegRNA MFE feature
            'MFE_RT-PBS-polyT'           : [round(fMFE3, 1)],
            'MFE_Spacer'                 : [round(fMFE4, 1)],

            # DeepSpCas9 score
            'DeepSpCas9_score'           : [SpCas9().predict([target[:30]]).SpCas9.loc[0]],
        }

        self.features = pd.DataFrame.from_dict(data=self.dict_feat, orient='columns')


    def predict(self, pe_system:str, cell_type:str = 'HEK293T', show_features:bool = False, report=False):

        df_all = self.features.copy()

        os.environ['CUDA_VISIBLE_DEVICES']='0'
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        
        model_info = LoadModel('DeepPrime', pe_system, cell_type)
        model_dir  = model_info.model_dir

        mean = pd.read_csv(f'{model_dir}/mean_231124.csv', header=None, index_col=0).squeeze()
        std  = pd.read_csv(f'{model_dir}/std_231124.csv',  header=None, index_col=0).squeeze()

        test_features = select_cols(df_all)

        g_test = seq_concat(df_all)
        x_test = (test_features - mean) / std

        g_test = torch.tensor(g_test, dtype=torch.float32, device=device)
        x_test = torch.tensor(x_test.to_numpy(), dtype=torch.float32, device=device)

        models = [m_files for m_files in glob(f'{model_dir}/*.pt')]
        preds  = []

        for m in models:
            model = GeneInteractionModel(hidden_size=128, num_layers=1).to(device)
            model.load_state_dict(torch.load(m, map_location=device))
            model.eval()
            with torch.no_grad():
                g, x = g_test, x_test
                g = g.permute((0, 3, 1, 2))
                pred = model(g, x).detach().cpu().numpy()
            preds.append(pred)
        
        # AVERAGE PREDICTIONS
        preds = np.squeeze(np.array(preds))
        preds = np.mean(preds, axis=0)
        preds = np.exp(preds) - 1

        df_all.insert(1, f'{pe_system}_score', preds)

        self.data = df_all

        return preds
    
    # def predict: END



def set_alt_position_window(sStrand, sAltKey, nAltIndex, nIndexStart, nIndexEnd, nAltLen):
    if sStrand == '+':

        if sAltKey.startswith('sub'):
            return (nAltIndex + 1) - (nIndexStart - 3)
        else:
            return (nAltIndex + 1) - (nIndexStart - 3)

    else:
        if sAltKey.startswith('sub'):
            return nIndexEnd - nAltIndex + 3 - (nAltLen - 1)

        elif sAltKey.startswith('del'):
            return nIndexEnd - nAltIndex + 3 - nAltLen

        else:
            return nIndexEnd - nAltIndex + 3 + nAltLen
        # if END:
    # if END:

# def END: set_alt_position_window


def set_PAM_nicking_pos(sStrand, sAltType, nAltLen, nAltIndex, nIndexStart, nIndexEnd):
    if sStrand == '-':
        nPAM_Nick = nIndexEnd + 3
    else:
        nPAM_Nick = nIndexStart - 3

    return nPAM_Nick

# def END: set_PAM_Nicking_Pos


def check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, sAltType, nAltLen, nAltIndex):
    nUp, nDown = dict_sWinSize[sAltType][nAltLen]

    if sStrand == '+':
        nPAMCheck_min = nAltIndex - nUp + 1
        nPAMCheck_max = nAltIndex + nDown + 1
    else:
        # nPAMCheck_min = nAltIndex - nDown + 1
        nPAMCheck_min = nAltIndex - nDown
        nPAMCheck_max = nAltIndex + nUp + 1
    # if END:

    if nIndexStart < nPAMCheck_min or nIndexEnd > nPAMCheck_max:
        return 0
    else:
        return 1

# def END: check_PAM_window

class PEFeatureExtraction:
    def __init__(self, input_params: dict):
        
        # configuration parameters
        # 이게 sRGN을 지정하는 단계까 .predict  부분이기 때문에, feature extraction 부분에서 spacer 길이를 조정할 수가 없음...!
        # pe_system을 지정하는 단계를 바꿔줘야 할 것 같음. 
        self.spacer_len = input_params['spacerlen']
        
        # Initialized
        self.get_input(input_params)
        self.get_sAltNotation(input_params['nAltIndex'])

        self.dict_sSeqs   = {}
        self.dict_sCombos = {}
        self.dict_sOutput = {}

        self.gc_cache   = {}
        self.tm_cache   = {}
        self.mfe_cache  = {}

    # def End: __init__
    
    def get_input(self, input_params: dict):
        self.sWTSeq     = input_params['wt_seq'].upper()
        self.sEditedSeq = input_params['ed_seq'].upper()
        self.sAltType   = input_params['edit_type']
        self.nAltLen    = input_params['edit_len']
        self.sAltKey    = self.sAltType + str(self.nAltLen)

        if   self.sAltType.startswith('sub'): self.type_sub = 1; self.type_del = 0; self.type_ins = 0
        elif self.sAltType.startswith('del'): self.type_del = 1; self.type_sub = 0; self.type_ins = 0
        elif self.sAltType.startswith('ins'): self.type_ins = 1; self.type_del = 0; self.type_sub = 0
    
    # def End: get_input

    def get_sAltNotation(self, nAltIndex):
        if self.sAltType == 'sub':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[nAltIndex:nAltIndex + self.nAltLen], self.sEditedSeq[nAltIndex:nAltIndex + self.nAltLen])

        elif self.sAltType == 'del':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[nAltIndex:nAltIndex + 1 + self.nAltLen], self.sEditedSeq[nAltIndex])

        else:
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[nAltIndex], self.sEditedSeq[nAltIndex:nAltIndex + self.nAltLen + 1])

    # def END: get_sAltNotation

    def get_all_RT_PBS(self, input_params:dict):
        """
        nMinPBS: If you set specific number, lower than MinPBS will be not generated. Default=0
        nMaxPBS: If you set specific number, higher than MinPBS will be not generated. Default=17
        nMaxRT = : If you set specific number, higher than MinPBS will be not generated. Default=40
        nSetPBSLen = 0  # Fix PBS Len: Set if >0
        nSetRTLen = 0  # Fix RT  Len: Set if >0
        PAM: 4-nt sequence
        """

        nAltIndex   = input_params['nAltIndex']
        pam         = input_params['pam']
        nMinPBS     = input_params['pbs_min']
        nMaxPBS     = input_params['pbs_max']
        nMaxRT      = input_params['rtt_max']
        nSetPBSLen  = 0
        nSetRTLen   = 0

        nMaxEditPosWin = nMaxRT + 3  # Distance between PAM and mutation

        dict_sWinSize = {'sub': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 2 - 3, 6], 3: [nMaxRT - 3 - 3, 6]},
                        'ins': {1: [nMaxRT - 2 - 3, 6], 2: [nMaxRT - 3 - 3, 6], 3: [nMaxRT - 4 - 3, 6]},
                        'del': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 1 - 3, 6], 3: [nMaxRT - 1 - 3, 6]}}

        
        if pam == 'NRCH': # for NRCH-PE PAM
            dict_sRE = {'+': '[ACGT][ACGT]G[ACGT]|[ACGT][CG]A[ACGT]|[ACGT][AG]CC|[ATCG]ATG', 
                        '-': '[ACGT]C[ACGT][ACGT]|[ACGT]T[CG][ACGT]|G[GT]T[ACGT]|ATT[ACGT]|CAT[ACGT]|GGC[ACGT]|GTA[ACGT]'} 
        elif pam == 'NGG':
            dict_sRE = {'+': '[ACGT]GG[ACGT]', '-': '[ACGT]CC[ACGT]'} # for Original-PE PAM
        elif pam == 'NAG':
            dict_sRE = {'+': '[ACGT]AG[ACGT]', '-': '[ACGT]CT[ACGT]'} # for Original-PE PAM
        elif pam == 'NGA':
            dict_sRE = {'+': '[ACGT]GA[ACGT]', '-': '[ACGT]TC[ACGT]'} # for Original-PE PAM
        elif pam == 'NNGG':
            dict_sRE = {'+': '[ACGT][ACGT]GG', '-': 'CC[ACGT][ACGT]'} # for sRGN-PE PAM

        for sStrand in ['+', '-']:

            sRE = dict_sRE[sStrand]
            for sReIndex in regex.finditer(sRE, self.sWTSeq, overlapped=True):

                if sStrand == '+':
                    nIndexStart = sReIndex.start()
                    nIndexEnd = sReIndex.end() - 1
                    sPAMSeq = self.sWTSeq[nIndexStart:nIndexEnd]
                    sGuideSeq = self.sWTSeq[nIndexStart - self.spacer_len:nIndexEnd]
                else:
                    nIndexStart = sReIndex.start() + 1
                    nIndexEnd = sReIndex.end()
                    sPAMSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd])
                    sGuideSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd + 20])

                nAltPosWin = set_alt_position_window(sStrand, self.sAltKey, nAltIndex, nIndexStart, nIndexEnd,
                                                    self.nAltLen)

                ## AltPosWin Filter ##
                if nAltPosWin <= 0:             continue
                if nAltPosWin > nMaxEditPosWin: continue

                nPAM_Nick = set_PAM_nicking_pos(sStrand, self.sAltType, self.nAltLen, nAltIndex, nIndexStart, nIndexEnd)

                if not check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, self.sAltType, self.nAltLen,
                                        nAltIndex): continue

                sPAMKey = '%s,%s,%s,%s,%s,%s,%s' % (
                    self.sAltKey, self.sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq)

                dict_sRT, dict_sPBS = self.determine_PBS_RT_seq(sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen,
                                                        nSetRTLen, nAltIndex, nPAM_Nick, nAltPosWin, self.sEditedSeq)

                nCnt1, nCnt2 = len(dict_sRT), len(dict_sPBS)
                if nCnt1 == 0: continue
                if nCnt2 == 0: continue
                
                if sPAMKey not in self.dict_sSeqs:
                    self.dict_sSeqs[sPAMKey] = ''
                self.dict_sSeqs[sPAMKey] = [dict_sRT, dict_sPBS]

            # loop END: sReIndex
        # loop END: sStrand


    # def END: get_all_RT_PBS

    
    def determine_PBS_RT_seq(self, sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen, nSetRTLen, nAltIndex, nPAM_Nick,
                            nAltPosWin, sForTempSeq):
        dict_sPBS = {}
        dict_sRT = {}

        list_nPBSLen = [nNo + 1 for nNo in range(nMinPBS, nMaxPBS)]
        for nPBSLen in list_nPBSLen:

            ## Set PBS Length ##
            if nSetPBSLen:
                if nPBSLen != nSetPBSLen: continue

            if sStrand == '+':
                nPBSStart = nPAM_Nick - nPBSLen  # 5' -> PamNick
                nPBSEnd = nPAM_Nick
                sPBSSeq = sForTempSeq[nPBSStart:nPBSEnd] # sForTempSeq = self.EditedSeq

            else:
                if self.sAltKey.startswith('sub'):
                    nPBSStart = nPAM_Nick
                elif self.sAltKey.startswith('ins'):
                    nPBSStart = nPAM_Nick + self.nAltLen
                elif self.sAltKey.startswith('del'):
                    nPBSStart = nPAM_Nick - self.nAltLen

                sPBSSeq = reverse_complement(sForTempSeq[nPBSStart:nPBSStart + nPBSLen]) # sForTempSeq = self.EditedSeq

            # if END: sStrand

            sKey = len(sPBSSeq)
            if sKey not in dict_sPBS:
                dict_sPBS[sKey] = ''
            dict_sPBS[sKey] = sPBSSeq
        # loop END: nPBSLen

        if sStrand == '+':
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo + 1 for nNo in range(nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            elif self.sAltKey.startswith('ins'):
                list_nRTPos = [nNo + 1 for nNo in range(nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            else:
                list_nRTPos = [nNo + 1 for nNo in range(nAltIndex, (nPAM_Nick + nMaxRT))] ## 수정! ## del2 RHA 3 del1 RHA2
        else:
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 1 - nMaxRT, nAltIndex)] ## 수정! ## sub1 sub 3 RHA 0
            else:
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 3 - nMaxRT, nAltIndex + self.nAltLen - 1)] ## 수정! ## ins2 최소가 2까지 ins3 RHA 최소 3 #del2 RHA 2 del1 RHA1
        for nRTPos in list_nRTPos:

            if sStrand == '+':
                nRTStart = nPAM_Nick  # PamNick -> 3'
                nRTEnd = nRTPos
                sRTSeq = sForTempSeq[nRTStart:nRTEnd]

            else:
                if self.sAltKey.startswith('sub'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick  # PamNick -> 3'
                elif self.sAltKey.startswith('ins'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick + self.nAltLen  # PamNick -> 3'
                elif self.sAltKey.startswith('del'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick - self.nAltLen  # PamNick -> 3'

                sRTSeq = reverse_complement(sForTempSeq[nRTStart:nRTEnd])

                if not sRTSeq: continue
            # if END: sStrand

            sKey = len(sRTSeq)

            ## Set RT Length ##
            if nSetRTLen:
                if sKey != nSetRTLen: continue

            ## Limit Max RT len ##
            if sKey > nMaxRT: continue

            ## min RT from nick site to mutation ##
            if self.sAltKey.startswith('sub'):
                if sStrand == '+':
                    if sKey < abs(nAltIndex - nPAM_Nick): continue
                else:
                    if sKey < abs(nAltIndex - nPAM_Nick + self.nAltLen - 1): continue ### 
            else:
                if sStrand == '-':
                    if sKey < abs(nAltIndex - nPAM_Nick + self.nAltLen - 1): continue

            if self.sAltKey.startswith('ins'):
                if sKey < nAltPosWin + 1: continue

            if sKey not in dict_sRT:
                dict_sRT[sKey] = ''
            dict_sRT[sKey] = sRTSeq
        # loop END: nRTPos

        return [dict_sRT, dict_sPBS]


    # def END: determine_PBS_RT_seq

    def make_rt_pbs_combinations(self):
        for sPAMKey in self.dict_sSeqs:

            dict_sRT, dict_sPBS = self.dict_sSeqs[sPAMKey]

            list_sRT = [dict_sRT[sKey] for sKey in dict_sRT]
            list_sPBS = [dict_sPBS[sKey] for sKey in dict_sPBS]

            if sPAMKey not in self.dict_sCombos:
                self.dict_sCombos[sPAMKey] = ''
            self.dict_sCombos[sPAMKey] = {'%s,%s' % (sRT, sPBS): {} for sRT in list_sRT for sPBS in list_sPBS}
        # loop END: sPAMKey


    # def END: make_rt_pbs_combinations


    def determine_seqs(self):
        for sPAMKey in self.dict_sSeqs:

            sAltKey, sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq = sPAMKey.split(',')
            nAltPosWin = int(nAltPosWin)
            nNickIndex = int(nPAM_Nick)

            for sSeqKey in self.dict_sCombos[sPAMKey]:

                sRTSeq, sPBSSeq = sSeqKey.split(',')

                ## for Tm1
                sForTm1 = reverse_complement_rna(sPBSSeq)

                if sStrand == '+':
                    ## for Tm2
                    sForTm2 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq)]
                    
                    ## for Tm3
                    if self.sAltType.startswith('sub'):
                        sForTm3 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq)]
                    elif self.sAltType.startswith('ins'):
                        sForTm3 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) - self.nAltLen]
                    else:  # del
                        sForTm3 = self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) + self.nAltLen]

                    ## for Tm4
                    if self.sAltType.startswith('sub'):
                        sTm4antiSeq = reverse_complement(self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq)])
                    elif self.sAltType.startswith('ins'):
                        sTm4antiSeq = reverse_complement(self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) - self.nAltLen])
                    else:  # del
                        sTm4antiSeq = reverse_complement(self.sWTSeq[nNickIndex:nNickIndex + len(sRTSeq) + self.nAltLen])                    

                else:
                    ## for Tm2
                    sForTm2 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq):nNickIndex])

                    ## for Tm3
                    if self.sAltType.startswith('sub'):
                        sForTm3 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq):nNickIndex])
                    elif self.sAltType.startswith('ins'):
                        sForTm3 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq) + self.nAltLen:nNickIndex])
                    else:  # del
                        sForTm3 = reverse_complement(self.sWTSeq[nNickIndex - len(sRTSeq) - self.nAltLen:nNickIndex])

                    ## for Tm4
                    if self.sAltType.startswith('sub'):
                        sTm4antiSeq = self.sWTSeq[nNickIndex - len(sRTSeq):nNickIndex]
                    elif self.sAltType.startswith('ins'):
                        sTm4antiSeq = self.sWTSeq[nNickIndex - len(sRTSeq) + self.nAltLen:nNickIndex]
                    else:  # del
                        sTm4antiSeq = self.sWTSeq[nNickIndex - len(sRTSeq) - self.nAltLen:nNickIndex]

                # if END

                sForTm4 = [sRTSeq, sTm4antiSeq]

                ## for Tm5
                # sForTm5 = [reverse_complement(sRTSeq.replace('A', 'U')), sRTSeq] # original code
                sForTm5 = reverse_complement_rna(sRTSeq)

                self.dict_sCombos[sPAMKey][sSeqKey] = {
                    'Tm1_PBS': sForTm1,
                    'Tm2_RTT_cTarget_sameLength': sForTm2,
                    'Tm3_RTT_cTarget_replaced': sForTm3,
                    'Tm4_cDNA_PAM-oppositeTarget': sForTm4,
                    'Tm5_RTT_cDNA': sForTm5
                }
            # loop END: sSeqKey
        # loop END: sPAMKey
    # def END: determine_seqs


    def determine_secondary_structure(self):
        for sPAMKey in self.dict_sSeqs:

            sAltKey, sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq = sPAMKey.split(',')
            list_sOutputKeys = ['Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
                                'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2', 'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS',
                                'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 'MFE_RT-PBS-polyT', 'MFE_Spacer']

            if sPAMKey not in self.dict_sOutput:
                self.dict_sOutput[sPAMKey] = {}

            sGuideGN19 = 'G' + sGuideSeq[1:-3] 

            for sSeqKey in self.dict_sCombos[sPAMKey]:

                if sSeqKey not in self.dict_sOutput[sPAMKey]:
                    self.dict_sOutput[sPAMKey][sSeqKey] = {sKey: '' for sKey in list_sOutputKeys}

                self.determine_Tm(sPAMKey, sSeqKey)
                self.determine_GC(sPAMKey, sSeqKey)
                self.determine_MFE(sPAMKey, sSeqKey, sGuideGN19)

            # loop END: sSeqKey

        # loop END: sPAMKey


    def determine_Tm(self, sPAMKey, sSeqKey):

        sequences = [
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm1_PBS'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm2_RTT_cTarget_sameLength'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm3_RTT_cTarget_replaced'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm4_cDNA_PAM-oppositeTarget'],
            self.dict_sCombos[sPAMKey][sSeqKey]['Tm5_RTT_cDNA'],
        ]
        
        ## Tm1 DNA/RNA mm1 ##
        fTm1 = mt.Tm_NN(seq=Seq(sequences[0]), nn_table=mt.R_DNA_NN1)

        ## Tm2 DNA/DNA mm0 ##
        fTm2 = mt.Tm_NN(seq=Seq(sequences[1]), nn_table=mt.DNA_NN3)

        ## Tm3 DNA/DNA mm0 ##
        fTm3 = mt.Tm_NN(seq=Seq(sequences[2]), nn_table=mt.DNA_NN3)

        ## Tm4 DNA/DNA mm1 ##
        if not sequences[3]: fTm4 = 0

        else:
            for sSeq1, sSeq2 in zip(sequences[3][0], sequences[3][1]):
                try: fTm4 = mt.Tm_NN(seq=sSeq1, c_seq=sSeq2, nn_table=mt.DNA_NN3)
                except ValueError: continue
            # loop END: sSeq1, sSeq2
        # if END:

        # Tm5 - revcom(AAGTcGATCC(RNA version)) + AAGTcGATCC
        fTm5 = mt.Tm_NN(seq=Seq(sequences[4]), nn_table=mt.R_DNA_NN1)

        # deltaTm (Tm3 - Tm2)
        delta_tm = fTm3 - fTm2

        # 결과 저장
        self.dict_sOutput[sPAMKey][sSeqKey].update({
            'Tm1_PBS'                    : fTm1,
            'Tm2_RTT_cTarget_sameLength' : fTm2,
            'Tm3_RTT_cTarget_replaced'   : fTm3,
            'Tm4_cDNA_PAM-oppositeTarget': fTm4,
            'Tm5_RTT_cDNA'               : fTm5,
            'deltaTm_Tm4-Tm2'            : delta_tm,
        })

    # def END: determine_Tm


    def determine_GC(self, sPAMKey, sSeqKey):
        sRTSeqAlt, sPBSSeq = sSeqKey.split(',')

        self.dict_sOutput[sPAMKey][sSeqKey].update({
            'GC_count_PBS'      : sPBSSeq.count('G') + sPBSSeq.count('C'),
            'GC_count_RTT'      : sRTSeqAlt.count('G') + sRTSeqAlt.count('C'),
            'GC_count_RT-PBS'   : (sPBSSeq + sRTSeqAlt).count('G') + (sPBSSeq + sRTSeqAlt).count('C'),
            'GC_contents_PBS'   : 100 * gc(sPBSSeq),
            'GC_contents_RTT'   : 100 * gc(sRTSeqAlt),
            'GC_contents_RT-PBS': 100 * gc(sPBSSeq + sRTSeqAlt)
        })

    # def END: determine_GC

    def determine_MFE(self, sPAMKey, sSeqKey, sGuideGN19):

        sRTSeq, sPBSSeq = sSeqKey.split(',')

        sequences = [
            reverse_complement(sPBSSeq + sRTSeq) + 'TTTTTT', # RT + PBS + PolyT
            sGuideGN19                                       # GN19 guide seq
        ]

        # MFE 계산
        mfe_values = [self.calculate_MFE(seq) for seq in sequences]

        # 결과 저장
        self.dict_sOutput[sPAMKey][sSeqKey].update({
            'MFE_RT-PBS-polyT': mfe_values[0],
            'MFE_Spacer'      : mfe_values[1],
        })

    # def END: determine_MFE

    def calculate_MFE(self, sequence):
        # 시퀀스가 캐시에 있는지 확인
        if sequence in self.mfe_cache:
            return self.mfe_cache[sequence]

        # 시퀀스가 캐시에 없을 경우 계산
        mfe_value = round(fold_compound(sequence).mfe()[1], 1)
        self.mfe_cache[sequence] = mfe_value  # 계산된 값을 캐시에 저장
        return mfe_value
    

    def make_output_df(self):

        list_output = []
        list_sOutputKeys = ['Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 'Tm4_cDNA_PAM-oppositeTarget', 
                            'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2', 'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS',
                            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 'MFE_RT-PBS-polyT', 'MFE_Spacer']

        for sPAMKey in self.dict_sSeqs:

            sAltKey, sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq = sPAMKey.split(',')
            nNickIndex = int(nPAM_Nick)

            if sStrand == '+':
                sWTSeq74 = self.sWTSeq[nNickIndex - 21:nNickIndex + 53]
                nEditPos = 61 - nNickIndex
            else:
                sWTSeq74 = reverse_complement(self.sWTSeq[nNickIndex - 53:nNickIndex + 21])
                if not self.sAltType.startswith('ins'):
                    nEditPos = nNickIndex - 60 - self.nAltLen + 1
                else:
                    nEditPos = nNickIndex - 59

            for sSeqKey in self.dict_sOutput[sPAMKey]:

                dict_seq = self.dict_sCombos[sPAMKey][sSeqKey]
                sRTTSeq, sPBSSeq = sSeqKey.split(',')
                PBSlen = len(sPBSSeq)
                RTlen = len(sRTTSeq)

                sPBS_RTSeq = sPBSSeq + sRTTSeq
                s5Bufferlen = 21 - PBSlen
                s3Bufferlen = 53 - RTlen
                sEDSeq74 = 'x' * s5Bufferlen + sPBS_RTSeq + 'x' * s3Bufferlen

                if self.sAltType.startswith('del'):
                    RHA_len = len(sRTTSeq) - nEditPos + 1
                else:
                    RHA_len = len(sRTTSeq) - nEditPos - self.nAltLen + 1


                list_sOut = [self.input_id, reverse_complement(sPBS_RTSeq), len(sPBSSeq), len(sRTTSeq), len(sPBSSeq + sRTTSeq), 
                            nEditPos, self.nAltLen,RHA_len, sWTSeq74, sEDSeq74, 
                            self.type_sub, self.type_ins, self.type_del
                            ] + [self.dict_sOutput[sPAMKey][sSeqKey][sKey] for sKey in list_sOutputKeys]

                list_output.append(list_sOut)
            
            # loop END: sSeqKey
        
        hder_features = [
            # Sample ID
            'ID', 

            # pegRNA sequence features
            'RT-PBS', 

            # pegRNA edit and length features
            'PBS_len', 'RTT_len', 'RT-PBS_len', 'Edit_pos', 'Edit_len', 'RHA_len', 
            
            # Target sequences
            'Target', 'Masked_EditSeq', 

            # Edit types
            'type_sub', 'type_ins', 'type_del',

            # Tm features
            'Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
            'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2',

            # GC counts and contents
            'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS', 
            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 
            
            # RNA 2ndary structure features
            'MFE_RT-PBS-polyT', 'MFE_Spacer'
            ]

        df_out = pd.DataFrame(list_output, columns=hder_features)

        list_spacer = [wt74[24-self.spacer_len:24] for wt74 in df_out.Target]
        df_out.insert(1, 'Spacer', list_spacer)
        
        # loop END: sPAMKey

        return df_out

# def END: make_output


class GeneInteractionModel(nn.Module):


    def __init__(self, hidden_size, num_layers, num_features=24, dropout=0.1):
        super(GeneInteractionModel, self).__init__()
        self.hidden_size = hidden_size
        self.num_layers = num_layers

        self.c1 = nn.Sequential(
            nn.Conv2d(in_channels=4, out_channels=128, kernel_size=(2, 3), stride=1, padding=(0, 1)),
            nn.BatchNorm2d(128),
            nn.GELU(),
        )
        self.c2 = nn.Sequential(
            nn.Conv1d(in_channels=128, out_channels=108, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(108),
            nn.GELU(),
            nn.AvgPool1d(kernel_size=2, stride=2),

            nn.Conv1d(in_channels=108, out_channels=108, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(108),
            nn.GELU(),
            nn.AvgPool1d(kernel_size=2, stride=2),

            nn.Conv1d(in_channels=108, out_channels=128, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm1d(128),
            nn.GELU(),
            nn.AvgPool1d(kernel_size=2, stride=2),
        )

        self.r = nn.GRU(128, hidden_size, num_layers, batch_first=True, bidirectional=True)

        self.s = nn.Linear(2 * hidden_size, 12, bias=False)

        self.d = nn.Sequential(
            nn.Linear(num_features, 96, bias=False),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(96, 64, bias=False),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(64, 128, bias=False)
        )

        self.head = nn.Sequential(
            nn.BatchNorm1d(140),
            nn.Dropout(dropout),
            nn.Linear(140, 1, bias=True),
        )

    def forward(self, g, x):
        g = torch.squeeze(self.c1(g), 2)
        g = self.c2(g)
        g, _ = self.r(torch.transpose(g, 1, 2))
        g = self.s(g[:, -1, :])

        x = self.d(x) ## 여기가 문제
        out = self.head(torch.cat((g, x), dim=1))
        return F.softplus(out)

def seq_concat(data, col1='Target', col2='Masked_EditSeq', seq_length=74):
    wt = preprocess_seq(data[col1], seq_length)
    # ed = preprocess_masked_seq(data[col2], seq_length)
    ed = preprocess_seq(data[col2], seq_length)
    g = np.concatenate((wt, ed), axis=1)
    g = 2 * g - 1

    return g


def select_cols(data):
    features = data.loc[:, ['PBS_len', 'RTT_len', 'RT-PBS_len', 'Edit_pos', 'Edit_len', 'RHA_len', 'type_sub',
                            'type_ins', 'type_del', 'Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
                            'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2',
                            'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS', 
                            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 
                            'MFE_RT-PBS-polyT', 'MFE_Spacer', 'DeepSpCas9_score']]

    return features



class DeepPrimeOff:
    def __init__(self):
        '''Pipeline for creating input for the DeepPrime-Off model and providing the model's output.
        
        ## How to use
        #### Step 1. Run Cas-OFFinder with the spacer sequence of pegRNAs

        #### Step 2. Setup the model
        ```python
        from genet.predict import DeepPrimeOff

        deep_off = DeepPrimeOff()
        deep_off.setup('./cas_offinder_results/cas')

        ```
        - cas_offinder_results: Path of text file with cas_offinder results.
        - fasta_path: Path of directory containing fasta files.
        - seq_length: Length of sequence context.

        #### Step 3. Predict the score

        ```python
        
        df_PE_off = deep_off.predict() # type: pd.DataFrame

        ```
        
        '''

        # Check Cas-OFFinder installed before importing this module
        # 만약 Cas-OFFinder가 아직 설치되어 있지 않다면, 현재 OS를 확인 후 적절한 binary file을 다운로드 한다. 


        




        pass
    
    # def END: __init__

    
    def setup(self,
              features:pd.DataFrame,
              cas_offinder_result:str,
              ref_genome:str='Homo sapiens', 
              download_fasta:bool=False,
              custom_genome:str=None, 
              ) -> pd.DataFrame:
        
        """일단 지금은 Cas-OFFinder output을 넣어주는 형태이지만, 
        나중에는 DeepPrime_record를 인식해서 spacer sequence를 뽑아내고, 자동으로 Cas-OFFinder가 돌아가게 만들기

        From cas_offinder_results, get on_target_scaper, Location, 
        Position, Off-target sequence, Strand, and number of mismatch (MM) information.

        Then, find and open the fasta file which matches the chromosome number of each sequence in cas_offinder_result.
        After then, find the sequence context starting from 'Position' to seq_length (74nt by default).
        This sequence contexts are returned as a DataFrame.

        Args:
            features (pd.DataFrame): _description_
            cas_offinder_result (str): Path of text file with cas_offinder results.
            ref_genome (str, optional): _description_. Defaults to 'Homo sapiens'.
            download_fasta (bool, optional): _description_. Defaults to False.
            custom_genome (str, optional): _description_. Defaults to None.
        
        Returns:
            pd.DataFrame: Input features for predicting off-target scores using the DeepPrime-Off model.
        """

        # Step1: Check if all required columns are present in the DeepPrime features DataFrame.
        self.features = self._check_record(features=features)

        # Step2: Check if there are FASTA files at the specified FASTA file path. 
        # If a path is specified for the custom genome (!= None), search for FASTA files at that path. 
        # self.fasta is the path (str) where FASTA files are stored.

        if custom_genome==None: 
            self.fasta = self._check_fasta(ref_genome=ref_genome, download_fasta=download_fasta)
        else:
            self.fasta = custom_genome

        # Step3: (TODO) Retrieve spacer sequences from self.features and the FASTA file path received from ref_genome.
        # After then, execute Cas-OFFinder.


        # Step4: Convert Cas-OFFinder result file to DataFrame format
        self.df_offinder = self._offinder_to_df(cas_offinder_result)

        # Step5: Retrieve the 74nt target context from the FASTA file.
        self.df_offinder = self._get_target_seq(df_offinder=self.df_offinder, ref_path=self.fasta)

        # Step6: Make the DeepPrime features DataFrame with off-target candidates for each pegRNA by combining them.
        self.features = self._match_target_seq(features=self.features, df_offinder=self.df_offinder)

        return self.features

    # def END: setup


    def predict(self, show_features:bool=False) -> pd.DataFrame:

        os.environ['CUDA_VISIBLE_DEVICES']='0'
        # df_all = self.features.copy()
        df_all = self.features

        data = df_all
        chunk_size = 10000

        chunks = [group for _, group in data.groupby(np.arange(len(data)) // chunk_size)]
        
        # make progress bar
        pbar = tqdm(chunks,
                    desc='DeepPrime-Off prediction',
                    total=len(chunks), 
                    unit=' M index', unit_scale=True, leave=False)
        
        # Combining np.ndarrays containing activated values for each index of chunked data
        preds = np.concatenate([self._model_worker(data=data) for data in pbar])
        
        zero_indices = []

        for i in range(len(df_all)):
            difference = 0
            on, ref = df_all['Target'].iloc[i], df_all['Off-context'].iloc[i]
            rt_len = df_all['RTT_len'].iloc[i]

            boundary = 17 + rt_len

            for j in range(boundary):
                if on[4+j] != ref[4+j]:
                    difference += 1
            
            if difference > 4:
                zero_indices.append(i)
        
        preds[zero_indices] = 0

        df_all.insert(1, f'DeepPrime-Off_score', preds)

        if   show_features == False: return df_all.iloc[:, :17]
        elif show_features == True : return df_all

    # def End: predict
        
    def _model_worker(self, data:pd.DataFrame) -> np.ndarray:
        """_summary_

        Args:
            data (pd.DataFrame): _description_

        Returns:
            np.ndarray: _description_
        """        

        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        
        model_info = LoadModel('DeepPrime', 'PE2-Off', 'HEK293T')
        model_dir  = model_info.model_dir

        mean = pd.read_csv(f'{model_dir}/mean_231124.csv', header=None, index_col=0).squeeze()
        std  = pd.read_csv(f'{model_dir}/std_231124.csv',  header=None, index_col=0).squeeze()

        test_features = select_cols(data)

        g_test = seq_concat(data, col1='Off-context', col2='Masked_EditSeq', seq_length=74)
        x_test = (test_features - mean) / std

        g_test = torch.tensor(g_test, dtype=torch.float32, device=device)
        x_test = torch.tensor(x_test.to_numpy(), dtype=torch.float32, device=device)

        models = [m_files for m_files in glob(f'{model_dir}/*.pt')]
        preds  = []

        for m in models:
            model = GeneInteractionModel(hidden_size=128, num_layers=1).to(device)
            model.load_state_dict(torch.load(m, map_location=device))
            model.eval()
            with torch.no_grad():
                g, x = g_test, x_test
                g = g.permute((0, 3, 1, 2))
                pred = model(g, x).detach().cpu().numpy()
            preds.append(pred)

        # AVERAGE PREDICTIONS
        preds = np.squeeze(np.array(preds))
        preds = np.mean(preds, axis=0)
        preds = np.exp(preds) - 1

        return preds

    # def End: _model_worker


    def _check_record(self, features:pd.DataFrame) -> pd.DataFrame:
        """A function that checks if the input features conform to the DataFrame format created by the DeepPrime pipeline.

        Args:
            features (pd.DataFrame): DataFrame containing information of the features required by DeepPrime.

        Raises:
            ValueError: If the column names that should be included in the features are missing, an error will occur.

        Returns:
            pd.DataFrame: Return the DataFrame containing features received as input without any modifications.
        """

        # Check if the input dataframe is in the correct format

        list_features = [
            'ID', 

            # pegRNA sequence features
            'Spacer', 'RT-PBS', 

            # pegRNA edit and length features
            'PBS_len', 'RTT_len', 'RT-PBS_len', 'Edit_pos', 'Edit_len', 'RHA_len', 
            
            # Target sequences
            'Target', 'Masked_EditSeq', 

            # Edit types
            'type_sub', 'type_ins', 'type_del',

            # Tm features
            'Tm1_PBS', 'Tm2_RTT_cTarget_sameLength', 'Tm3_RTT_cTarget_replaced', 
            'Tm4_cDNA_PAM-oppositeTarget', 'Tm5_RTT_cDNA', 'deltaTm_Tm4-Tm2',

            # GC counts and contents
            'GC_count_PBS', 'GC_count_RTT', 'GC_count_RT-PBS', 
            'GC_contents_PBS', 'GC_contents_RTT', 'GC_contents_RT-PBS', 
            
            # RNA 2ndary structure features
            'MFE_RT-PBS-polyT', 'MFE_Spacer',

            # DeepSpCas9 score
            'DeepSpCas9_score',
            ]
        
        for feat_name in list_features:
            if feat_name not in features.columns:
                raise ValueError(f'The input dataframe does not have the column {feat_name}')
            
        return features
    
    # def END: _check_record
            
    def _check_fasta(self, ref_genome:str, download_fasta:bool) -> str:
        """Try to locate the chromosomes using GetChromosome from the path of ref_genome.
        If the species cannot be found based on chromosomes, attempt to locate them using GetGenome.
        If none of the files are found, raise a NotFoundError. 
        However, if download_fasta is set to True, download the fasta files.

        Args:
            ref_genome (str): The path where the FASTA file is stored or the path where it will be downloaded.
            download_fasta (bool): Whether to download the FASTA file from the FTP server if it is not found in the specified path.

        Returns:
            str: The path where the reference FASTA file is stored.
        """

        ref_path   = ref_genome.replace('\\', '/')
        ref_genome = os.path.basename(ref_path)

        genome_meta = GetGenome(ref_genome)

        # Case1: If the scaffold level is "chromosome," then check if the files in GetChromosome.contents() are included in ref_path.
        if genome_meta.info['assembly_level'] == 'Chromosome':

            chrom = GetChromosome(id=ref_genome)

            for f in chrom.contents():

                if os.path.isfile(os.path.join(ref_path, f)): continue
                else:
                    if download_fasta == True: chrom.download(f, download_path=ref_path)
                    else: raise FileNotFoundError(f'FASTA file {f} not found in "{ref_path}" directory. If you want to download it automatically, please set download_fasta=True.')

        # Case2: If the scaffold level is not "chromosome," then try to find it in GetGenome.contents().
        else:
            for f in genome_meta.contents():

                # In genome_meta, download the file that ends with '_genomic.fna.gz' rather than downloading the entire contents.
                if not f.endswith('_genomic.fna.gz'): continue
                
                if os.path.isfile(os.path.join(ref_path, f)): break
                else:
                    if download_fasta == True: 
                        genome_meta.download(f, download_path=ref_path)
                        break
                    else: 
                        raise FileNotFoundError(f'FASTA file {f} not found in "{ref_path}" directory. If you want to download it automatically, please set download_fasta=True.')
            
        return ref_path
    # def END: _check_fasta
        

    def _offinder_to_df(self, cas_offinder_result_path:str) -> pd.DataFrame:
        """For cas_offinder_result, transform tsv to DataFrame format.
        Also, add "Chromosome" column.

        Args:
            cas_offinder_result_path (str): The path of original text file of Cas-OFFinder results

        Returns:
            pd.DataFrame: DataFrame formatted Cas-OFFinder result
        """

        df_offinder = pd.read_csv(cas_offinder_result_path, sep='\t', names=['On_target_scaper', 'Location', 'Position', 'Off_target_sequence', 'Strand', 'MM_count'])

        # extract chromosome name from 'Location' and make new column 'Chromosome'
        df_offinder['Chromosome'] = df_offinder['Location'].apply(lambda x: x.split(' ')[0])

        return df_offinder
    # def END: _offinder_to_df


    def _get_target_seq(self, df_offinder:pd.DataFrame, ref_path:str='Homo sapiens') -> pd.DataFrame:
        """From FASTA file, get sequence context starting from 'Position' to seq_length (default 74nt).
        This function searches for the FASTA file with the name received through ref_path.
        The FASTA file should match the one referenced during Cas-OFFinder execution.

        Args:
            df_offinder (pd.DataFrame): DataFrame formatted Cas-OFFinder result
            ref_path (str, optional): The path where the FASTA file is stored. Defaults to 'Homo sapiens'.

        Returns:
            pd.DataFrame: DataFrame with the 74nt sequence context of the off-targets found by Cas-OFFinder.
        """
        
        seq_length = 74
        
        list_df_out = []
        df_offinder_grouped = df_offinder.groupby('Chromosome')

        # make progress bar
        pbar = tqdm(df_offinder_grouped.groups.keys(),
                    desc='Finding sequence context',
                    total=len(df_offinder_grouped.groups.keys()), 
                    unit=' chromosomes', unit_scale=True, leave=False)

        # iterate through each chromosome
        for chromosome in pbar:
            
            # fasta  = str(SeqIO.read(f'{ref_path}/chr{chromosome}.fna', 'fasta').seq)
            file_name = f'chr{chromosome}'

            fasta  = str(self._open_fasta_record(file_name, ref_path=ref_path).seq)
            df_chr = df_offinder_grouped.get_group(chromosome)

            chr_strand_grouped = df_chr.groupby('Strand')

            # for strand == '+'
            df_strand_fwd = chr_strand_grouped.get_group('+').copy()
            df_strand_fwd['Off74_context'] = df_strand_fwd['Position'].apply(lambda pos: fasta[pos-4:pos-4+seq_length])
            list_df_out.append(df_strand_fwd)

            # for strand == '-'
            df_strand_rev = chr_strand_grouped.get_group('-').copy()
            df_strand_rev['Off74_context'] = df_strand_rev['Position'].apply(lambda pos: reverse_complement(fasta[pos+28-seq_length:pos+28]))
            list_df_out.append(df_strand_rev)

        return pd.concat(list_df_out, axis=0)
    # def END: _get_target_seq
    
    def _open_fasta_record(self, file_name:str, ref_path:str='Homo sapiens') -> str:
        """Check if there is a file with the specified 'file_name' in the 'ref_path' directory, regardless of the file extension. 
        If there is a file with one of the following extensions: .fa, .fna, .fasta, .fa.gz, .fna.gz, or .fasta.gz, 
        return the file path and name.

        Args:
            file_name (str): The filename to be opened as a Seq record within the given directory.
            ref_path (str, optional): The path where the reference genome FASTA file is stored. Defaults to 'Homo sapiens'.

        Returns:
            SeqRecord: Parsing information of the FASTA file.
        """

        files = glob(f'{ref_path}/{file_name}.*')

        if len(files) == 0:
            raise FileNotFoundError(f'The FASTA file {file_name} does not exist in {ref_path}')
        
        else:
            for file in files:
                if file.endswith('.fa') or file.endswith('.fna') or file.endswith('.fasta'):
                    return SeqIO.read(file, 'fasta')
                
                elif file.endswith('.fa.gz') or file.endswith('.fna.gz') or file.endswith('.fasta.gz'):
                    return SeqIO.read(gzip.open(file, 'rt'), 'fasta')
                
            else:
                raise FileNotFoundError(f'The FASTA file {file_name} does not exist in {ref_path}')
    
    # def END: _open_fasta_record


    def _match_target_seq(self, features:pd.DataFrame, df_offinder:pd.DataFrame) -> pd.DataFrame:
        """DeepPrime pipeline에서 만들어진 features record에 off-target candidates로 찾아진 74nt sequence를 연결해주는 함수.
        이때, 

        Args:
            features (pd.DataFrame): DeepPrime pipeline으로부터 만들어진 pegRNA / target sequence와 biofeatures가 있는 DataFrame.
            df_offinder (pd.DataFrame): Cas-OFFinder result DataFrame.

        Returns:
            pd.DataFrame: Features record에 off-target candidates 74nt sequence를 매칭시킨 DataFrame.
        """

        # ToDo?: 먼저 index가 중복되지 않은 것이 들어있는지 확인한다. 만약 중복 index가 있으면, 새 index를 배정한다. 
        # 하지만 DeepPrime pipeline에서 있던 dataframe을 그대로 가져왔다면, 중복 index는 없을 것... 

        list_df    = []
        feat_group = features.groupby('Spacer')

        # make progress bar
        pbar = tqdm(df_offinder.index,
                    desc='Make DeepPrime-Off input DataFrame',
                    total=len(df_offinder.index),
                    unit=' Off-targets', unit_scale=True, leave=False)

        for idx_off in pbar:
            df_off_row = df_offinder.loc[idx_off]

            spacer_on  = df_off_row['On_target_scaper'][:-3]
            location   = df_off_row['Location']
            position   = df_off_row['Position']
            off_target = df_off_row['Off_target_sequence']
            strand     = df_off_row['Strand']
            n_mismatch = df_off_row['MM_count']
            off74seq   = df_off_row['Off74_context']
            
            try:
                df_feat_temp = feat_group.get_group(spacer_on).copy()
                len_feat = len(df_feat_temp)

                df_feat_temp.insert(10, f'MM_num',      [n_mismatch for _ in range(len_feat)])
                df_feat_temp.insert(10, f'Strand',      [strand     for _ in range(len_feat)])
                df_feat_temp.insert(10, f'Position',    [position   for _ in range(len_feat)])
                df_feat_temp.insert(10, f'Location',    [location   for _ in range(len_feat)])
                df_feat_temp.insert(10, f'Off-context', [off74seq   for _ in range(len_feat)])
                df_feat_temp.insert(10, f'Off-target',  [off_target for _ in range(len_feat)])
            
            except: continue

            list_df.append(df_feat_temp)
            
        df_out = pd.concat(list_df, ignore_index=True)

        return df_out
    # def END: _match_target_seq












